# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import re

import pytest

pytest.importorskip("dask")
import numpy as np
from dask import array as da
from numpy.testing import assert_allclose

from astropy import units as u
from astropy.modeling.fitting import (
    FitInfoArrayContainer,
    LevMarLSQFitter,
    TRFLSQFitter,
    parallel_fit_dask,
)
from astropy.modeling.models import (
    Const1D,
    Gaussian1D,
    Linear1D,
    Planar2D,
)
from astropy.nddata import NDData, StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.compat.optional_deps import HAS_PLT
from astropy.wcs import WCS


def gaussian(x, amplitude, mean, stddev):
    return amplitude * np.exp(-((x - mean) ** 2) / 2 / stddev**2)


def modify_axes(array, shape, swapaxes):
    array = array.reshape(shape)
    if swapaxes is not None:
        array = array.swapaxes(*swapaxes)
    return array


@pytest.mark.parametrize(
    "iterating_shape,swapaxes_data,swapaxes_parameters,fitting_axes",
    [
        ((120,), None, None, 0),
        ((2, 3, 4, 5), None, None, 0),
        ((2, 3, 4, 5), (0, 1), None, 1),
        ((2, 3, 4, 5), (0, 2), (0, 1), 2),
    ],
)
def test_1d_model_fit_axes(
    iterating_shape, swapaxes_data, swapaxes_parameters, fitting_axes
):
    # Test that different iterating and fitting axes work. We start off by
    # creating an array with one fitting and one iterating dimension and we then
    # modify the array dimensionality and order of the dimensions to test
    # different scenarios.

    N = 120
    P = 20

    rng = np.random.default_rng(12345)

    x = np.linspace(-5, 30, P)

    amplitude = rng.uniform(1, 10, N)
    mean = rng.uniform(0, 25, N)
    stddev = rng.uniform(1, 4, N)

    data = gaussian(x[:, None], amplitude, mean, stddev)

    # At this point, the data has shape (P, N)

    # Modify the axes by reshaping and swapping axes
    data = modify_axes(data, data.shape[0:1] + iterating_shape, swapaxes_data)

    # Set initial parameters to be close to but not exactly equal to true parameters

    model = Gaussian1D(
        amplitude=modify_axes(
            amplitude * rng.random(N), iterating_shape, swapaxes_parameters
        ),
        mean=modify_axes(mean + rng.random(N), iterating_shape, swapaxes_parameters),
        stddev=modify_axes(
            stddev + rng.random(N), iterating_shape, swapaxes_parameters
        ),
    )
    fitter = LevMarLSQFitter()

    model_fit = parallel_fit_dask(
        data=data,
        model=model,
        fitter=fitter,
        fitting_axes=fitting_axes,
        world=(x,),
        scheduler="synchronous",
    )

    # Check that shape and values match

    assert_allclose(
        model_fit.amplitude.value,
        modify_axes(amplitude, iterating_shape, swapaxes_parameters),
    )
    assert_allclose(
        model_fit.mean.value, modify_axes(mean, iterating_shape, swapaxes_parameters)
    )
    assert_allclose(
        model_fit.stddev.value,
        modify_axes(stddev, iterating_shape, swapaxes_parameters),
    )


@pytest.mark.parametrize(
    "iterating_shape,swapaxes_data,swapaxes_parameters,fitting_axes",
    [
        ((120,), None, None, (0, 1)),
        ((2, 3, 4, 5), None, None, (0, 1)),  # make iterating dimensions N-d
        ((2, 3, 4, 5), (0, 1), None, (1, 0)),  # swap data axes to be (y, x)
        ((2, 3, 4, 5), (1, 2), None, (0, 2)),  # start mixing up axes
        ((2, 3, 4, 5), (0, 3), (0, 1), (3, 1)),  # iterating axes out of order
    ],
)
def test_2d_model_fit_axes(
    iterating_shape, swapaxes_data, swapaxes_parameters, fitting_axes
):
    N = 120
    P = 20
    Q = 10

    rng = np.random.default_rng(12345)

    x = np.linspace(-5, 30, P)
    y = np.linspace(5, 20, Q)

    slope_x = rng.uniform(1, 5, N)
    slope_y = rng.uniform(1, 5, N)
    intercept = rng.uniform(-2, 2, N)

    data = slope_x * x[:, None, None] + slope_y * y[None, :, None] + intercept

    # At this point, the data has shape (P, Q, N)

    # Modify the axes by reshaping and swapping axes
    data = modify_axes(data, data.shape[0:2] + iterating_shape, swapaxes_data)

    # Set initial parameters to be close to but not exactly equal to true parameters

    model = Planar2D(
        slope_x=modify_axes(
            slope_x * rng.random(N), iterating_shape, swapaxes_parameters
        ),
        slope_y=modify_axes(
            slope_y + rng.random(N), iterating_shape, swapaxes_parameters
        ),
        intercept=modify_axes(
            intercept + rng.random(N), iterating_shape, swapaxes_parameters
        ),
    )
    fitter = LevMarLSQFitter()

    model_fit = parallel_fit_dask(
        data=data,
        model=model,
        fitter=fitter,
        fitting_axes=fitting_axes,
        world=(x, y),
        scheduler="synchronous",
    )

    # Check that shape and values match

    assert_allclose(
        model_fit.slope_x.value,
        modify_axes(slope_x, iterating_shape, swapaxes_parameters),
    )
    assert_allclose(
        model_fit.slope_y.value,
        modify_axes(slope_y, iterating_shape, swapaxes_parameters),
    )
    assert_allclose(
        model_fit.intercept.value,
        modify_axes(intercept, iterating_shape, swapaxes_parameters),
    )


class TestWorld:
    def test_no_world(self):
        # This also doubles as a test when there are no iterating dimensions
        data = gaussian(np.arange(20), 2, 10, 1)
        model = Gaussian1D(amplitude=1.5, mean=12, stddev=1.5)
        fitter = LevMarLSQFitter()
        model_fit = parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=0,
            scheduler="synchronous",
        )
        assert_allclose(model_fit.amplitude.value, 2)
        assert_allclose(model_fit.mean.value, 10)
        assert_allclose(model_fit.stddev.value, 1)

    def test_wcs_1d(self):
        # Test specifying world as a WCS, for the 1D model case

        data = gaussian(
            np.arange(20)[:, None],
            np.array([2, 1.8]),
            np.array([10, 11]),
            np.array([1, 1.1]),
        )
        model = Gaussian1D(amplitude=1.5, mean=1.2, stddev=0.15)
        fitter = LevMarLSQFitter()

        wcs = WCS(naxis=2)
        wcs.wcs.ctype = "OFFSET", "WAVE"
        wcs.wcs.crval = 10, 0.1
        wcs.wcs.crpix = 1, 1
        wcs.wcs.cdelt = 10, 0.1

        model_fit = parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=0,
            world=wcs,
            scheduler="synchronous",
        )
        assert_allclose(model_fit.amplitude.value, [2, 1.8])
        assert_allclose(model_fit.mean.value, [1.1, 1.2])
        assert_allclose(model_fit.stddev.value, [0.1, 0.11])

    def test_wcs_pixel_dimension_mismatch(self):
        data = np.empty((20, 10))
        model = Gaussian1D(amplitude=1.5, mean=1.2, stddev=0.15)
        fitter = LevMarLSQFitter()

        wcs = WCS(naxis=3)

        with pytest.raises(
            ValueError,
            match=re.escape(
                "The WCS pixel_n_dim (3) does not match the number of "
                "dimensions in the data (2)"
            ),
        ):
            parallel_fit_dask(
                data=data,
                model=model,
                fitter=fitter,
                fitting_axes=0,
                world=wcs,
                scheduler="synchronous",
            )

    def test_wcs_world_dimension_mismatch(self):
        data = np.empty((20, 10))
        model = Gaussian1D(amplitude=1.5, mean=1.2, stddev=0.15)
        fitter = LevMarLSQFitter()

        wcs = WCS(naxis=2)
        wcs.wcs.ctype = "RA---TAN", "DEC--TAN"
        wcs.wcs.crval = 10.0, 20.0
        wcs.wcs.crpix = 1, 1
        wcs.wcs.cdelt = 0.01, 0.01

        with pytest.raises(
            ValueError,
            match=re.escape(
                "The number of WCS world axes corresponding to the fitting axes "
                "(2) does not match the number of fitting axes (1)"
            ),
        ):
            parallel_fit_dask(
                data=data,
                model=model,
                fitter=fitter,
                fitting_axes=0,
                world=wcs,
                scheduler="synchronous",
            )

    def test_array(self):
        # Test specifying world as a tuple of arrays with dimensions matching the data

        data = gaussian(
            np.arange(21)[:, None],
            np.array([2, 1.8]),
            np.array([10, 10]),
            np.array([1, 1.1]),
        )
        model = Gaussian1D(amplitude=1.5, mean=7, stddev=2)
        fitter = LevMarLSQFitter()

        world1 = np.array([np.linspace(0, 10, 21), np.linspace(5, 15, 21)]).T

        model_fit = parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=0,
            world=(world1,),
            scheduler="synchronous",
        )
        assert_allclose(model_fit.amplitude.value, [2, 1.8])
        assert_allclose(model_fit.mean.value, [5, 10])
        assert_allclose(model_fit.stddev.value, [0.5, 0.55])

    def test_array_length_mismatch(self):
        data = np.empty((20, 2))
        model = Gaussian1D(amplitude=1.5, mean=7, stddev=2)
        fitter = LevMarLSQFitter()

        world1 = np.array([np.linspace(0, 10, 21), np.linspace(5, 15, 21)]).T

        with pytest.raises(
            ValueError,
            match=re.escape(
                "The number of world arrays (2) must match number of fitting axes (1)"
            ),
        ):
            parallel_fit_dask(
                data=data,
                model=model,
                fitter=fitter,
                fitting_axes=0,
                world=(world1, world1),
            )

    def test_array_shape_mismatch(self):
        data = np.empty((20, 2))
        model = Gaussian1D(amplitude=1.5, mean=7, stddev=2)
        fitter = LevMarLSQFitter()

        world1 = np.array([np.linspace(0, 10, 31), np.linspace(5, 15, 31)]).T

        with pytest.raises(
            ValueError,
            match=re.escape(
                "The arrays in the world tuple should be broadcastable to the "
                "shape of the data (expected (20, 2)), got (31, 2))"
            ),
        ):
            parallel_fit_dask(
                data=data,
                model=model,
                fitter=fitter,
                fitting_axes=0,
                world=(world1,),
            )

    def test_array_dimension_mismatch(self):
        model = Planar2D()
        data = np.empty((20, 10, 5))
        fitter = LevMarLSQFitter()

        with pytest.raises(
            ValueError,
            match=re.escape(
                "world[2] has length 6 but data along dimension 2 has length 5"
            ),
        ):
            parallel_fit_dask(
                data=data,
                model=model,
                fitter=fitter,
                fitting_axes=(1, 2),
                world=(np.arange(10), np.arange(6)),
            )

    def test_invalid_type(self):
        data = np.empty((20, 2))
        model = Gaussian1D(amplitude=1.5, mean=7, stddev=2)
        fitter = LevMarLSQFitter()

        world1 = np.array([np.linspace(0, 10, 21), np.linspace(5, 15, 21)]).T

        with pytest.raises(
            TypeError,
            match=re.escape("world should be None, a WCS object or a tuple of arrays"),
        ):
            parallel_fit_dask(
                data=data,
                model=model,
                fitter=fitter,
                fitting_axes=0,
                world="banana",
            )


def test_fitter_kwargs(tmp_path):
    data = gaussian(np.arange(20), 2, 10, 1)
    data = np.broadcast_to(data.reshape((20, 1)), (20, 3)).copy()

    data[0, 0] = np.nan

    model = Gaussian1D(amplitude=1.5, mean=12, stddev=1.5)
    fitter = LevMarLSQFitter()

    model_fit = parallel_fit_dask(
        data=data,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        fitter_kwargs={"filter_non_finite": True},
        scheduler="synchronous",
    )

    assert_allclose(model_fit.amplitude.value, 2)
    assert_allclose(model_fit.mean.value, 10)
    assert_allclose(model_fit.stddev.value, 1)


class TestDiagnostics:
    def setup_method(self, method):
        self.data = gaussian(np.arange(20), 2, 10, 1)
        self.data = np.broadcast_to(self.data.reshape((20, 1)), (20, 3)).copy()
        self.data_original = self.data.copy()
        self.data[0, 0] = np.nan
        self.model = Gaussian1D(amplitude=1.5, mean=12, stddev=1.5)
        self.fitter = LevMarLSQFitter()

    def test_error(self, tmp_path):
        parallel_fit_dask(
            data=self.data,
            model=self.model,
            fitter=self.fitter,
            fitting_axes=0,
            diagnostics="error",
            diagnostics_path=tmp_path / "diag1",
            scheduler="synchronous",
        )

        assert os.listdir(tmp_path / "diag1") == ["0"]
        assert sorted(os.listdir(tmp_path / "diag1" / "0")) == ["error.log"]

    def test_all(self, tmp_path):
        parallel_fit_dask(
            data=self.data,
            model=self.model,
            fitter=self.fitter,
            fitting_axes=0,
            diagnostics="all",
            diagnostics_path=tmp_path / "diag2",
            scheduler="synchronous",
        )

        assert sorted(os.listdir(tmp_path / "diag2")) == ["0", "1", "2"]

    def test_all_world_wcs(self, tmp_path):
        # Make sure things world also with world=wcs

        parallel_fit_dask(
            data=self.data,
            model=self.model,
            fitter=self.fitter,
            fitting_axes=0,
            world=WCS(naxis=2),
            diagnostics="error",
            diagnostics_path=tmp_path / "diag3",
            scheduler="synchronous",
        )

        assert os.listdir(tmp_path / "diag3") == ["0"]
        assert sorted(os.listdir(tmp_path / "diag3" / "0")) == ["error.log"]

    @pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib.pyplot")
    def test_callable(self, tmp_path):
        # And check that we can pass in a callable

        def custom_callable(path, world, data, weights, model, fitting_kwargs):
            import matplotlib.pyplot as plt

            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.plot(world[0], data, "k.")
            ax.text(0.1, 0.9, "Fit failed!", color="r", transform=ax.transAxes)
            fig.savefig(os.path.join(path, "fit.png"))
            plt.close(fig)

        # Note: here we keep the default scheduler ('processes') to make sure
        # that callables are passed correctly to other processes. This test is
        # not as fast as other ones anyway due to the plotting so this doesn't
        # have a big performance impact.

        parallel_fit_dask(
            data=self.data,
            model=self.model,
            fitter=self.fitter,
            fitting_axes=0,
            world=WCS(naxis=2),
            diagnostics="error",
            diagnostics_path=tmp_path / "diag4",
            diagnostics_callable=custom_callable,
        )

        assert os.listdir(tmp_path / "diag4") == ["0"]
        assert sorted(os.listdir(tmp_path / "diag4" / "0")) == ["error.log", "fit.png"]

    def test_warnings(self, tmp_path):
        # Check that catching warnings works

        parallel_fit_dask(
            data=self.data_original,
            model=self.model,
            fitter=self.fitter,
            fitting_axes=0,
            diagnostics="error+warn",
            diagnostics_path=tmp_path / "diag5",
            fitter_kwargs={"maxiter": 2},
            scheduler="synchronous",
        )

        assert sorted(os.listdir(tmp_path / "diag5")) == ["0", "1", "2"]
        assert sorted(os.listdir(tmp_path / "diag5" / "0")) == ["warn.log"]

    def test_missing_path(self):
        with pytest.raises(ValueError, match="diagnostics_path should be set"):
            parallel_fit_dask(
                data=self.data,
                model=self.model,
                fitter=self.fitter,
                fitting_axes=0,
                diagnostics="error",
            )

    def test_invalid(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "diagnostics should be None, 'error', 'error+warn', or 'all'"
            ),
        ):
            parallel_fit_dask(
                data=self.data,
                model=self.model,
                fitter=self.fitter,
                fitting_axes=0,
                diagnostics="spam",
            )


@pytest.mark.parametrize(
    "scheduler", (None, "synchronous", "processes", "threads", "default")
)
def test_dask_scheduler(scheduler):
    N = 120
    P = 20

    rng = np.random.default_rng(12345)

    x = np.linspace(-5, 30, P)

    amplitude = rng.uniform(1, 10, N)
    mean = rng.uniform(0, 25, N)
    stddev = rng.uniform(1, 4, N)

    data = gaussian(x[:, None], amplitude, mean, stddev)

    # At this point, the data has shape (P, N)
    # Set initial parameters to be close to but not exactly equal to true parameters

    model = Gaussian1D(
        amplitude=amplitude * rng.random(N),
        mean=mean + rng.random(N),
        stddev=stddev + rng.random(N),
    )
    fitter = LevMarLSQFitter()

    model_fit = parallel_fit_dask(
        data=data,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        world=(x,),
        scheduler=scheduler,
    )

    # Check that shape and values match

    assert_allclose(model_fit.amplitude.value, amplitude)
    assert_allclose(model_fit.mean.value, mean)
    assert_allclose(model_fit.stddev.value, stddev)


def test_compound_model():
    # Compound models have to be treated a little differently so check they
    # work fine.

    data = gaussian(
        np.arange(20)[:, None],
        np.array([2, 1.8]),
        np.array([10, 11]),
        np.array([1, 1.1]),
    )

    data[:, 0] += 2
    data[:, 1] += 3

    model1 = Gaussian1D(amplitude=1.5, mean=1.2, stddev=0.15)
    model2 = Const1D(1)

    model = model1 + model2

    fitter = LevMarLSQFitter()

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "OFFSET", "WAVE"
    wcs.wcs.crval = 10, 0.1
    wcs.wcs.crpix = 1, 1
    wcs.wcs.cdelt = 10, 0.1

    model_fit = parallel_fit_dask(
        data=data,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        world=wcs,
        scheduler="synchronous",
    )
    assert_allclose(model_fit.amplitude_0.value, [2, 1.8])
    assert_allclose(model_fit.mean_0.value, [1.1, 1.2])
    assert_allclose(model_fit.stddev_0.value, [0.1, 0.11])
    assert_allclose(model_fit.amplitude_1.value, [2, 3])

    # Check that constraints work

    model.amplitude_1 = 2
    model.amplitude_1.fixed = True

    model_fit = parallel_fit_dask(
        data=data,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        world=wcs,
        scheduler="synchronous",
    )

    assert_allclose(model_fit.amplitude_0.value, [2, 1.633], atol=0.001)
    assert_allclose(model_fit.mean_0.value, [1.1, 1.145], atol=0.001)
    assert_allclose(model_fit.stddev_0.value, [0.1, 0.736], atol=0.001)
    assert_allclose(model_fit.amplitude_1.value, [2, 2], atol=0.001)


def test_model_dimension_mismatch():
    model = Planar2D()
    data = np.empty((20, 10, 5))
    fitter = LevMarLSQFitter()

    with pytest.raises(
        ValueError,
        match=re.escape("Model is 2-dimensional, but got 1 value(s) in fitting_axes="),
    ):
        parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=0,
        )

    model = Linear1D()
    data = np.empty((20, 10, 5))
    fitter = LevMarLSQFitter()

    with pytest.raises(
        ValueError,
        match=re.escape("Model is 1-dimensional, but got 2 value(s) in fitting_axes="),
    ):
        parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=(1, 2),
        )


def test_data_dimension_mismatch():
    model = Planar2D()
    data = np.empty((20, 10, 5))
    fitter = LevMarLSQFitter()

    with pytest.raises(
        ValueError,
        match=re.escape("Fitting index 4 out of range for 3-dimensional data"),
    ):
        parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=(1, 4),
        )


class TestDaskInput:
    # Check that things work correctly when passing dask arrays as input.

    def setup_method(self, method):
        self.base_data = gaussian(np.arange(20), 2, 10, 1)
        self.base_data = np.broadcast_to(
            self.base_data.reshape((20, 1)), (20, 3)
        ).copy()
        self.model = Gaussian1D(amplitude=1.5, mean=12, stddev=1.5)
        self.fitter = LevMarLSQFitter()

    @pytest.mark.parametrize("preserve_native_chunks", (False, True))
    def test_data(self, preserve_native_chunks):
        model_fit = parallel_fit_dask(
            data=da.from_array(self.base_data, chunks=(20, 1)),
            model=self.model,
            fitter=self.fitter,
            fitting_axes=0,
            preserve_native_chunks=preserve_native_chunks,
            scheduler="synchronous",
        )

        assert_allclose(model_fit.amplitude.value, 2)
        assert_allclose(model_fit.mean.value, 10)
        assert_allclose(model_fit.stddev.value, 1)

    @pytest.mark.parametrize("preserve_native_chunks", (False, True))
    def test_data_and_weights(self, preserve_native_chunks):
        data = self.base_data.copy()
        data[0, 0] = 1000.0
        data = da.from_array(data, chunks=(20, 1))

        weights = (data < 100).astype(float)

        model_fit = parallel_fit_dask(
            data=data,
            weights=weights,
            model=self.model,
            fitter=self.fitter,
            fitting_axes=0,
            preserve_native_chunks=preserve_native_chunks,
            scheduler="synchronous",
        )

        assert_allclose(model_fit.amplitude.value, 2)
        assert_allclose(model_fit.mean.value, 10)
        assert_allclose(model_fit.stddev.value, 1)

    def test_preserve_native_chunks_invalid_data_chunks(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "When using preserve_native_chunks=True, the chunk size should match the data size along the fitting axe"
            ),
        ):
            parallel_fit_dask(
                data=da.from_array(self.base_data, chunks=(5, 2)),
                model=self.model,
                fitter=self.fitter,
                fitting_axes=0,
                preserve_native_chunks=True,
            )

    def test_preserve_native_chunks_invalid_weight_chunks(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "When using preserve_native_chunks=True, the weights should have the same chunk size as the data"
            ),
        ):
            parallel_fit_dask(
                data=da.from_array(self.base_data, chunks=(20, 1)),
                weights=da.random.random(self.base_data.shape).rechunk((18, 2)),
                model=self.model,
                fitter=self.fitter,
                fitting_axes=0,
                preserve_native_chunks=True,
            )

    def test_preserve_native_chunks_invalid_input_data_type(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "Can only set preserve_native_chunks=True if input data is a dask array"
            ),
        ):
            parallel_fit_dask(
                data=self.base_data,
                model=self.model,
                fitter=self.fitter,
                fitting_axes=0,
                preserve_native_chunks=True,
            )

    def test_preserve_native_chunks_invalid_input_weights_type(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "Can only set preserve_native_chunks=True if input weights is a dask array (if specified)"
            ),
        ):
            parallel_fit_dask(
                data=da.from_array(self.base_data, chunks=(20, 1)),
                weights=np.random.random(self.base_data.shape),
                model=self.model,
                fitter=self.fitter,
                fitting_axes=0,
                preserve_native_chunks=True,
            )


def test_weights():
    # Test specifying world as a tuple of arrays with dimensions matching the data

    data = gaussian(
        np.arange(21)[:, None],
        np.array([2, 1.8]),
        np.array([5, 10]),
        np.array([1, 1.1]),
    )

    # Introduce outliers but then adjust weights to ignore them
    data[10, 0] = 1000.0
    data[20, 1] = 2000.0
    weights = (data < 100.0).astype(float)

    model = Gaussian1D(amplitude=1.5, mean=7, stddev=2)
    fitter = LevMarLSQFitter()

    model_fit = parallel_fit_dask(
        data=data,
        weights=weights,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        scheduler="synchronous",
    )
    assert_allclose(model_fit.amplitude.value, [2, 1.8])
    assert_allclose(model_fit.mean.value, [5, 10])
    assert_allclose(model_fit.stddev.value, [1.0, 1.1])


class TestUnits:
    def test_basic(self):
        # Make sure that fitting with units works

        data = (
            gaussian(
                np.arange(21)[:, None],
                np.array([2, 1.8]),
                np.array([5, 10]),
                np.array([1, 1.1]),
            )
            * u.Jy
        )

        model = Gaussian1D(amplitude=1.5 * u.Jy, mean=7 * u.um, stddev=0.002 * u.mm)
        fitter = LevMarLSQFitter()

        model_fit = parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=0,
            world=(1000 * np.arange(21) * u.nm,),
            scheduler="synchronous",
        )

        assert_quantity_allclose(model_fit.amplitude.quantity, [2, 1.8] * u.Jy)
        assert_quantity_allclose(model_fit.mean.quantity, [5, 10] * u.um)
        assert_quantity_allclose(model_fit.stddev.quantity, [1.0, 1.1] * u.um)

    def test_units_no_input_units(self):
        # Make sure that fitting with units works for models without input_units defined

        data = (np.repeat(3, 20)).reshape((20, 1)) * u.Jy

        model = Const1D(1 * u.mJy)
        fitter = LevMarLSQFitter()

        assert not model.input_units

        model_fit = parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=0,
            world=(1000 * np.arange(20) * u.nm,),
            scheduler="synchronous",
        )

        assert_quantity_allclose(model_fit.amplitude.quantity, 3 * u.Jy)

    def test_units_with_wcs(self):
        data = gaussian(np.arange(20), 2, 10, 1).reshape((20, 1)) * u.Jy
        model = Gaussian1D(amplitude=1.5 * u.Jy, mean=7 * u.um, stddev=0.002 * u.mm)
        fitter = LevMarLSQFitter()

        wcs = WCS(naxis=2)
        wcs.wcs.ctype = "OFFSET", "WAVE"
        wcs.wcs.crval = 10, 0.1
        wcs.wcs.crpix = 1, 1
        wcs.wcs.cdelt = 10, 0.1
        wcs.wcs.cunit = "deg", "um"

        model_fit = parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=0,
            world=wcs,
            scheduler="synchronous",
        )
        assert_allclose(model_fit.amplitude.quantity, 2 * u.Jy)
        assert_allclose(model_fit.mean.quantity, 1.1 * u.um)
        assert_allclose(model_fit.stddev.quantity, 0.1 * u.um)

    def test_units_with_wcs_2d(self):
        # Fit a 2D model with mixed units to a dataset

        N = 3
        P = 20
        Q = 10

        wcs = WCS(naxis=3)
        wcs.wcs.ctype = "OFFSET", "WEIGHTS", "WAVE"
        wcs.wcs.crval = 10, 0.1, 0.2
        wcs.wcs.crpix = 1, 1, 1
        wcs.wcs.cdelt = 10, 0.1, 0.2
        wcs.wcs.cunit = "deg", "kg", "mm"

        rng = np.random.default_rng(12345)

        x = wcs.pixel_to_world(0, 0, np.arange(P))[2]
        y = wcs.pixel_to_world(0, np.arange(Q), 0)[1]

        slope_x = [1, 3, 2] * u.Jy / u.mm
        slope_y = [-1, 2, 3] * u.Jy / u.kg
        intercept = [5, 6, 7] * u.Jy

        data = slope_x * x[:, None, None] + slope_y * y[None, :, None] + intercept

        # At this point, the data has shape (P, Q, N)

        model = Planar2D(
            slope_x=slope_x * rng.uniform(0.9, 1.1, N),
            slope_y=slope_y * rng.uniform(0.9, 1.1, N),
            intercept=intercept * rng.uniform(0.9, 1.1, N),
        )
        fitter = LevMarLSQFitter()

        model_fit = parallel_fit_dask(
            data=data,
            model=model,
            fitter=fitter,
            fitting_axes=(0, 1),
            world=wcs,
            scheduler="synchronous",
        )
        assert_allclose(model_fit.slope_x.quantity, slope_x)
        assert_allclose(model_fit.slope_y.quantity, slope_y)
        assert_allclose(model_fit.intercept.quantity, intercept)


def test_skip_empty_data(tmp_path):
    # Test when one of the datasets being fit is all NaN

    data = gaussian(
        np.arange(21)[:, None],
        np.array([2, 1.8]),
        np.array([5, 10]),
        np.array([1, 1.1]),
    )

    data[:, 1] = np.nan

    model = Gaussian1D(amplitude=1.5, mean=7, stddev=2)
    fitter = TRFLSQFitter()

    model_fit = parallel_fit_dask(
        data=data,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        diagnostics="error+warn",
        diagnostics_path=tmp_path,
        scheduler="synchronous",
    )

    # If we don't properly skip empty data sections, then the fitting would fail
    # and populate the diagnostics directory.
    assert len(sorted(tmp_path.iterdir())) == 0

    assert_allclose(model_fit.amplitude.value, [2, np.nan])
    assert_allclose(model_fit.mean.value, [5, np.nan])
    assert_allclose(model_fit.stddev.value, [1.0, np.nan])


def test_world_wcs_axis_correlation():
    # Regression test for a bug that caused the world coordinates to not be
    # properly extracted from a WCS object if the axis correlation matrix
    # resulted in a difference in order between pixel and world coordinates.

    model = Gaussian1D()
    fitter = TRFLSQFitter()
    data = gaussian(
        np.arange(1, 6)[:, None],
        np.array([5, 5]),
        np.array([3, 2]),
        np.array([1, 1]),
    ).T

    common_kwargs = dict(data=data, model=model, fitter=fitter, scheduler="synchronous")

    # First case: a simple WCS - as the fitting axis is 1 in Numpy order, this
    # means we should use the world coordinates for the first WCS dimension. In
    # this case, the Gaussian means should be 3 and 2.

    wcs1 = WCS(naxis=2)
    wcs1.wcs.cdelt = 1, 2

    model_fit = parallel_fit_dask(fitting_axes=1, world=wcs1, **common_kwargs)
    assert_allclose(model_fit.mean, [3, 2])

    # Second case: as above, but WCS axes swapped. In this case, the means
    # should be 6 and 4.

    wcs1 = WCS(naxis=2)
    wcs1.wcs.cdelt = 2, 1

    model_fit = parallel_fit_dask(fitting_axes=1, world=wcs1, **common_kwargs)
    assert_allclose(model_fit.mean, [6, 4])

    # Third case: as in first case, but this time we set the PC matrix such
    # that the world axes are in a different order to their corresponding pixel
    # axis. In this case, the means should be 6 and 4 because fitting_axes=1
    # should correspond to the second WCS dimension.

    wcs3 = WCS(naxis=2)
    wcs3.wcs.cdelt = 1, 2
    wcs3.wcs.pc = [[0, 1], [1, 0]]

    model_fit = parallel_fit_dask(fitting_axes=1, world=wcs1, **common_kwargs)
    assert_allclose(model_fit.mean, [6, 4])


def test_support_nddata():
    data = gaussian(np.arange(20), 2, 10, 1).reshape((20, 1)) * u.Jy

    # Introduce outliers
    data[10, 0] = 1000.0 * u.Jy
    # Mask the outliers (invalid is True)
    mask = data > 100.0 * u.Jy

    model = Gaussian1D(amplitude=1.5 * u.Jy, mean=7 * u.um, stddev=0.002 * u.mm)
    fitter = LevMarLSQFitter()

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "OFFSET", "WAVE"
    wcs.wcs.crval = 10, 0.1
    wcs.wcs.crpix = 1, 1
    wcs.wcs.cdelt = 10, 0.1
    wcs.wcs.cunit = "deg", "um"

    nd_data = NDData(
        data=data,
        wcs=wcs,
        mask=mask,
    )

    model_fit = parallel_fit_dask(
        data=nd_data,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        scheduler="synchronous",
    )

    assert_allclose(model_fit.amplitude.quantity, 2 * u.Jy)
    assert_allclose(model_fit.mean.quantity, 1.1 * u.um)
    assert_allclose(model_fit.stddev.quantity, 0.1 * u.um)


def test_support_nddata_uncert():
    data = np.repeat(np.array([1, 2, 4]), 2).reshape((2, -1), order="F").T
    uncert = (
        np.repeat(np.array([1 / 7**0.5, 1 / 2**0.5, 1]), 2)
        .reshape((2, -1), order="F")
        .T
    )

    model = Const1D(0)
    fitter = LevMarLSQFitter()

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "OFFSET", "WAVE"
    wcs.wcs.crval = 10, 1
    wcs.wcs.crpix = 1, 1
    wcs.wcs.cdelt = 10, 1
    wcs.wcs.cunit = "deg", "m"

    nd_data = NDData(
        data=data,
        wcs=wcs,
        uncertainty=StdDevUncertainty(uncert),
    )

    model_fit = parallel_fit_dask(
        data=nd_data,
        model=model,
        fitter=fitter,
        fitting_axes=0,
        scheduler="synchronous",
    )

    assert_allclose(model_fit.amplitude, 1.5)


class TestFitInfo:
    def setup_method(self, method):
        self.data = gaussian(np.arange(20), 2, 10, 1)
        self.data = np.broadcast_to(self.data.reshape((20, 1)), (20, 3)).copy()
        self.data_original = self.data.copy()
        self.data[0, 0] = np.nan
        self.model = Gaussian1D(amplitude=1.5, mean=12, stddev=1.5)

    def test_default(self, tmp_path):
        fitter = TRFLSQFitter()

        parallel_fit_dask(
            data=self.data,
            model=self.model,
            fitter=fitter,
            fitting_axes=0,
            scheduler="synchronous",
        )

        assert fitter.fit_info is None

    def test_all(self, tmp_path):
        fitter = TRFLSQFitter()

        parallel_fit_dask(
            data=self.data,
            model=self.model,
            fitter=fitter,
            fitting_axes=0,
            scheduler="synchronous",
            fit_info=True,
        )

        assert "message" in fitter.fit_info.properties

        assert_allclose(fitter.fit_info.get_property_as_array("nfev"), [0, 9, 9])

        param_cov_array = fitter.fit_info.get_property_as_array("param_cov")
        assert param_cov_array.shape == (3, 3, 3)
        assert_allclose(param_cov_array[0], 0)
        assert_allclose(param_cov_array[1], param_cov_array[2])
        assert np.any(np.abs(param_cov_array[1]) > 0)

        # Test slicing that returns an array

        assert fitter.fit_info.shape == (3,)
        fit_info_subset = fitter.fit_info[:2]
        assert isinstance(fit_info_subset, FitInfoArrayContainer)
        assert fit_info_subset.shape == (2,)
        assert_allclose(fit_info_subset.get_property_as_array("nfev"), [0, 9])

        # Test slicing that returns a one element array

        fit_info_subset_single = fitter.fit_info[1:2]
        assert isinstance(fit_info_subset_single, FitInfoArrayContainer)
        assert fit_info_subset_single.shape == (1,)
        assert_allclose(fit_info_subset_single.get_property_as_array("nfev"), [9])

        # Test slicing that returns a scalar

        fit_info_indiv = fitter.fit_info[1]
        assert not isinstance(fit_info_indiv, FitInfoArrayContainer)
        assert fit_info_indiv.nfev == 9
        assert fit_info_indiv.message != ""

    def test_subset(self, tmp_path):
        fitter = TRFLSQFitter()

        parallel_fit_dask(
            data=self.data,
            model=self.model,
            fitter=fitter,
            fitting_axes=0,
            scheduler="synchronous",
            fit_info=("message", "nfev", "success"),
        )

        assert fitter.fit_info.properties == ("message", "nfev", "success")

        assert_allclose(fitter.fit_info.get_property_as_array("nfev"), [0, 9, 9])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import io

import pytest

from astropy.coordinates import FK4, FK5, ICRS, AltAz, Galactic, Supergalactic
from astropy.io.votable.exceptions import E16, W27, W57
from astropy.io.votable.table import parse
from astropy.io.votable.tree import CooSys
from astropy.utils.data import get_pkg_data_filename

COOSYS_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
<VOTABLE xmlns="http://www.ivoa.net/xml/VOTable/v{}" version="{}">
  <RESOURCE>
    {}
  </RESOURCE>
</VOTABLE>
"""
COOSYS_REFPOSITION = (
    '<COOSYS ID="my_coosys" epoch="J2015.5" system="ICRS" refposition="BARYCENTER"/>'
)
COOSYS_NO_REFPOSITION = '<COOSYS ID="my_coosys" epoch="J2015.5" system="ICRS"/>'

COOSYS_SYSTEM_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
<VOTABLE xmlns="http://www.ivoa.net/xml/VOTable/v1.3" version="{}">
  <RESOURCE>
    <COOSYS ID="my_coosys" epoch="J2015.5" system="{}"/>
  </RESOURCE>
</VOTABLE>
"""


def test_accept_refposition():
    for ns_version, version, coosys, expected_warnings in [
        ("1.1", "1.1", COOSYS_NO_REFPOSITION, None),
        ("1.2", "1.2", COOSYS_NO_REFPOSITION, W27),
        ("1.3", "1.3", COOSYS_NO_REFPOSITION, None),
        ("1.3", "1.4", COOSYS_NO_REFPOSITION, None),
        ("1.3", "1.5", COOSYS_NO_REFPOSITION, None),
        ("1.1", "1.1", COOSYS_REFPOSITION, W57),
        ("1.2", "1.2", COOSYS_REFPOSITION, (W27, W57)),
        ("1.3", "1.3", COOSYS_REFPOSITION, W57),
        ("1.3", "1.4", COOSYS_REFPOSITION, W57),
        ("1.3", "1.5", COOSYS_REFPOSITION, None),
    ]:
        xml_string = COOSYS_TEMPLATE.format(ns_version, version, coosys)
        input = io.BytesIO(bytes(xml_string, "utf-8"))
        if expected_warnings is not None:
            with pytest.warns(expected_warnings) as record:
                vot = parse(input, verify="warn")
            # If expected_warnings is a tuple, pytest.warns checked for at least one of the warnings,
            # so make sure we got the right number of them.
            if isinstance(expected_warnings, tuple):
                assert len(record) == len(expected_warnings)
        else:
            vot = parse(input, verify="warn")


def test_coosys_system():
    for version, system, expected_warning in [
        ("1.4", "ICRS", None),
        ("1.4", "supergalactic", None),
        ("1.4", "InvalidSystem", E16),
        ("1.5", "ICRS", None),
        ("1.5", "supergalactic", None),
        ("1.5", "InvalidSystem", E16),
    ]:
        xml_string = COOSYS_SYSTEM_TEMPLATE.format(version, system)
        input = io.BytesIO(bytes(xml_string, "utf-8"))
        if expected_warning is not None:
            with pytest.warns(expected_warning):
                parse(input, verify="warn")
        else:
            parse(input, verify="warn")


def _coosys_tests(votable):
    assert len(list(votable.iter_coosys())) == 3
    coosys = votable.get_coosys_by_id("coosys0")
    assert coosys.system == "ICRS"
    assert coosys.equinox == "J2000"
    assert coosys.epoch == "J2014.5"
    assert coosys.refposition == "BARYCENTER"

    coosys = votable.get_coosys_by_id("coosys1")
    assert coosys.system == "ICRS"
    assert coosys.equinox == "J2000"
    assert coosys.epoch == "J2015.5"
    assert coosys.refposition == "BARYCENTER"

    coosys = votable.get_coosys_by_id("coosys2")
    assert coosys.system == "FK4"
    assert coosys.equinox == "B1950"
    assert coosys.epoch == "B1950.0"
    assert coosys.refposition == "UNKNOWN"


def test_coosys():
    votable = parse(get_pkg_data_filename("data/coosys.xml"))
    _coosys_tests(votable)


def test_coosys_roundtrip():
    orig_votable = parse(get_pkg_data_filename("data/coosys.xml"))
    bio = io.BytesIO()
    orig_votable.to_xml(bio)
    bio.seek(0)
    votable = parse(bio)
    _coosys_tests(votable)


test_to_astropy_frame = [
    ("ICRS", None, None, ICRS()),
    ("FK4", "B1950", "B1950", FK4(equinox="B1950")),
    ("FK5", "J2000", None, FK5(equinox="J2000")),
    ("eq_FK4", "B1950", "B1950", FK4(equinox="B1950")),
    ("eq_FK5", "J2000", None, FK5(equinox="J2000")),
    ("GALACTIC", None, None, Galactic()),
    ("galactic", None, None, Galactic()),
    ("SUPER_GALACTIC", None, None, Supergalactic()),
    ("supergalactic", None, None, Supergalactic()),
    ("AZ_EL", None, "J2020", AltAz()),
]


@pytest.mark.parametrize("system,equinox,epoch,expected", test_to_astropy_frame)
def test_coosys_to_astropy_frame_and_time(system, equinox, epoch, expected):
    coosys = CooSys(equinox=equinox, epoch=epoch, system=system)
    assert coosys.to_astropy_frame() == expected


def test_coosys_to_astropy_frame_error():
    with pytest.raises(
        ValueError,
        match=(
            "There is no direct correspondence between 'BODY' and an astropy frame.*"
        ),
    ):
        CooSys(system="BODY").to_astropy_frame()



import traceback
import warnings
from copy import deepcopy
from math import ceil, log10, prod
from pathlib import Path

import numpy as np

import astropy.units as u
from astropy.modeling.utils import _combine_equivalency_dict
from astropy.nddata import NDUncertainty, StdDevUncertainty, support_nddata
from astropy.wcs.wcsapi import BaseHighLevelWCS, BaseLowLevelWCS
from astropy.wcs.wcsapi.wrappers import SlicedLowLevelWCS

__all__ = ["FitInfoArrayContainer", "parallel_fit_dask"]


def _pixel_to_world_values_block(*pixel, wcs=None):
    """
    Convert a block of pixel values to world values using a WCS. This is for
    use in map_blocks.
    """
    world = wcs.array_index_to_world_values(*pixel)
    world = np.array(world)
    return world


def _wcs_to_world_dask(wcs, shape, chunks=None):
    """
    Given a WCS and a data shape, return an iterable of dask arrays
    representing the world coordinates of the array.
    """
    import dask.array as da

    pixel = tuple([np.arange(size) for size in shape])
    pixel_nd = da.meshgrid(*pixel, indexing="ij")
    world = da.map_blocks(
        _pixel_to_world_values_block,
        *pixel_nd,
        wcs=deepcopy(wcs),
        new_axis=0,
        chunks=(len(shape),) + chunks,
    )
    return tuple([world[idx] for idx in range(len(shape))])


def _copy_with_new_parameters(model, parameters):
    # Make a copy of the model, setting the parameters to new values
    model_new = model.copy()
    model_new._reset_parameters(**parameters)
    return model_new


class FitInfoArrayContainer:
    """
    This class is intended to contain the object array of all fit_info values
    and provide a convenience method to access specific items from fit_info
    as arrays.
    """

    def __init__(self, fit_info_array):
        self._fit_info_array = fit_info_array

    @property
    def shape(self):
        return self._fit_info_array.shape

    @property
    def ndim(self):
        return self._fit_info_array.ndim

    def __getitem__(self, item):
        result = self._fit_info_array[item]
        if hasattr(result, "ndim"):
            return FitInfoArrayContainer(result)
        else:
            return result

    def get_property_as_array(self, name):
        """
        Return an array of one of the fit information properties

        Parameters
        ----------
        name : str
            The name of a property present on the individual fit information
            objects.
        """
        array = None
        for index in np.ndindex(self._fit_info_array.shape):
            fit_info = self._fit_info_array[index]
            if fit_info is not None:
                value = np.array(getattr(fit_info, name))
                if array is None:
                    array = np.zeros(self.shape + value.shape, dtype=value.dtype)
                if value.shape != array.shape[self.ndim :]:
                    raise ValueError(
                        "Property {name} does not have consistent shape in all fit_info"
                    )
                array[index] = value
        return array

    @property
    def properties(self):
        """
        The properties available to query with :meth:`~astropy.modeling.fitting.FitInfoArrayContainer.get_property_as_array`
        """
        # Find the first non-None .fit_info
        for index in np.ndindex(self._fit_info_array.shape):
            fit_info = self._fit_info_array[index]
            if fit_info is not None:
                return tuple(sorted(fit_info))
        return ()


class FitInfoSubset(dict):
    def __getattr__(self, attr):
        return self[attr]


def _fit_models_to_chunk(
    data,
    *arrays,
    block_info=None,
    model=None,
    fitter=None,
    world=None,
    diagnostics=None,
    diagnostics_path=None,
    diagnostics_callable=None,
    iterating_shape=None,
    fitter_kwargs=None,
    iterating_axes=None,
    fitting_axes=None,
    weights_specified=None,
    fit_info=None,
):
    """
    Function that gets passed to map_blocks and will fit models to a specific
    chunk of the data.
    """
    if fitter_kwargs is None:
        fitter_kwargs = {}

    # Start off by re-ordering axes so that iterating axes come first followed
    # by fitting axes
    original_axes = tuple(idx for idx in (iterating_axes + fitting_axes))
    new_axes = tuple(range(data.ndim))
    data = np.moveaxis(data, original_axes, new_axes)
    arrays = [np.moveaxis(array, original_axes, new_axes) for array in arrays]

    if weights_specified:
        weights = arrays[0]
        arrays = arrays[1:]
    else:
        weights = None

    # World coordinates can be specified either as Nd world arrays (in which
    # case the world kwarg is set to `None`), or passed in via the world kwarg
    # (if the world coordinates are given as 1D arrays)
    if world is None:
        parameters = arrays[: -model.n_inputs]
        world_arrays = arrays[-model.n_inputs :]
    else:
        parameters = arrays

    # Make the parameters into an Nd array, as this is what we will return. We
    # then modify this array in-place in the rest of the function.
    parameters = np.array(parameters)

    # In some cases, dask calls this function with empty arrays, so we can
    # take a short-cut here.
    if data.ndim == 0 or data.size == 0 or block_info is None or block_info == []:
        return parameters

    # Because of the way map_blocks works, we need to have all arrays passed
    # to map_blocks have the same shape, even though for the parameters this
    # means there are extra unneeded dimensions. We slice these out here.
    index = tuple([slice(None)] * (1 + len(iterating_axes)) + [0] * len(fitting_axes))
    parameters = parameters[index]

    # Transform array to object array and add one more index along the first
    # dimension so that we can store the fit_info
    if fit_info:
        parameters = parameters.astype(object)
        parameters = np.pad(parameters, [(0, 1)] + [(0, 0)] * (parameters.ndim - 1))

    # The world argument is used to pass through 1D arrays of world coordinates
    # (otherwise world_arrays is used) so if the model has more than one
    # dimension we need to make these arrays N-dimensional.
    if world is not None:
        if model.n_inputs > 1:
            world_values = np.meshgrid(*world, indexing="ij")
        else:
            world_values = world

    iterating_shape_chunk = data.shape[: len(iterating_axes)]

    model_i = model.copy()

    for index in np.ndindex(iterating_shape_chunk):
        # If all data values are NaN, just set parameters to NaN and move on
        if np.all(np.isnan(data[index])):
            for ipar in range(len(model.param_names)):
                parameters[(ipar,) + index] = np.nan
            continue

        # Inject parameters into model
        model_i._reset_parameters(
            **{
                name: parameters[(ipar,) + index]
                for ipar, name in enumerate(model.param_names)
            },
        )

        output = diagnostics == "all"
        error = ""
        all_warnings = []

        if world is None:
            world_values = tuple([w[index] for w in world_arrays])

        if weights is None:
            weights_kwargs = {}
        else:
            weights_kwargs = dict(weights=weights[index])

        # Do the actual fitting - note that we can use inplace=True here to
        # speed things up by avoiding an unnecessary copy, since we don't need
        # to retain the original parameter values.
        try:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                model_fit = fitter(
                    model_i,
                    *world_values,
                    data[index],
                    inplace=True,
                    **weights_kwargs,
                    **fitter_kwargs,
                )
                all_warnings.extend(w)
        except Exception as exc:
            model_fit = None
            if diagnostics is not None and diagnostics.startswith("error"):
                output = True
            error = traceback.format_exc()
            for ipar in range(len(model_i.param_names)):
                parameters[(ipar,) + index] = np.nan

            parameters[(-1,) + index] = None
        else:
            # Put fitted parameters back into parameters arrays. These arrays are
            # created in-memory by dask and are local to this process so should be
            # safe to modify in-place
            for ipar, name in enumerate(model_fit.param_names):
                parameters[(ipar,) + index] = getattr(model_fit, name).value

            if fit_info is True:
                parameters[(-1,) + index] = fitter.fit_info
            elif fit_info:
                fit_info_dict = {}
                for key in fit_info:
                    if hasattr(fitter.fit_info, key):
                        fit_info_dict[key] = getattr(fitter.fit_info, key)
                    else:
                        raise AttributeError(
                            f"fit_info on fitter has no attribute '{key}'"
                        )
                parameters[(-1,) + index] = FitInfoSubset(fit_info_dict)

        if diagnostics == "error+warn" and len(all_warnings) > 0:
            output = True

        if output:
            # Construct a folder name based on the iterating index. Currently i
            # i a 1-d index but we need to re-convert it back to an N-dimensional
            # index.

            index_abs = np.array(index) + np.array(
                [block_info[0]["array-location"][idx][0] for idx in iterating_axes]
            )
            maxlen = ceil(log10(max(iterating_shape)))
            fmt = "{0:0" + str(maxlen) + "d}"
            index_folder = Path(diagnostics_path).joinpath(
                "_".join(fmt.format(idx) for idx in index_abs)
            )
            index_folder.mkdir(parents=True, exist_ok=True)

            # Output error, if any
            if error:
                index_folder.joinpath("error.log").write_text(error)

            if all_warnings:
                index_folder.joinpath("warn.log").write_text(
                    "".join(f"{warning}\n" for warning in all_warnings)
                )

            if diagnostics_callable is not None:
                diagnostics_callable(
                    index_folder,
                    world_values,
                    data[index],
                    None if weights is None else weights[index],
                    model_fit,
                    fitter_kwargs,
                )

    return parameters


class ParameterContainer:
    """
    This is an array container intended to be passed to dask's ``from_array``.

    The initial parameter values need to be broadcast up to the data shape so
    that map_blocks can then iterate over both the data and parameters. We
    need to control the final chunking so that it matches the data. However,
    rather than use dask to do the broadcasting and rechunking, which creates a
    complex graph and results in high memory usage, this class can be used
    instead to do all the broadcasting on-the-fly with Numpy as needed and keeps
    the dask graph simple.
    """

    def __init__(self, values, iterating_shape, iterating_axes, data_shape):
        self._values = values
        self.shape = data_shape
        self.ndim = len(data_shape)
        self.dtype = float
        self._iterating_shape = iterating_shape
        self._iterating_axes = iterating_axes

    def __getitem__(self, item):
        values = np.broadcast_to(self._values, self._iterating_shape)
        values = values.reshape(
            tuple(
                self.shape[idx] if idx in self._iterating_axes else 1
                for idx in range(self.ndim)
            )
        )
        values = np.broadcast_to(values, self.shape)
        return values[item]


@support_nddata(wcs="world", uncertainty="weights", unit="data_unit")
def parallel_fit_dask(
    *,
    model,
    fitter,
    data,
    data_unit=None,
    weights=None,
    mask=None,
    fitting_axes=None,
    world=None,
    chunk_n_max=None,
    diagnostics=None,
    diagnostics_path=None,
    diagnostics_callable=None,
    scheduler=None,
    fitter_kwargs=None,
    preserve_native_chunks=False,
    equivalencies=None,
    fit_info=False,
):
    """
    Fit a model in parallel to an N-dimensional dataset.

    Axes in the N-dimensional dataset are considered to be either 'fitting
    axes' or 'iterating axes'. As a specific example, if fitting a
    spectral cube with two celestial and one spectral axis, then if fitting a
    1D model to each spectrum in the cube, the spectral axis would be a fitting
    axis and the celestial axes would be iterating axes.

    Parameters
    ----------
    model : :class:`astropy.modeling.Model`
        The model to fit, specifying the initial parameter values. The shape
        of the parameters should be broadcastable to the shape of the iterating
        axes.
    fitter : :class:`astropy.modeling.fitting.Fitter`
        The fitter to use in the fitting process.
    data : `numpy.ndarray` or `dask.array.core.Array`
        The N-dimensional data to fit.
    data_units : `astropy.units.Unit`
        Units for the data array, for when the data array is not a ``Quantity``
        instance.
    weights : `numpy.ndarray`, `dask.array.core.Array` or `astropy.nddata.NDUncertainty`
        The weights to use in the fitting. See the documentation for specific
        fitters for more information about the meaning of weights.
        If passed as a `.NDUncertainty` object it will be converted to a
        `.StdDevUncertainty` and then passed to the fitter as 1 over that.
    mask : `numpy.ndarray`
        A boolean mask to be applied to the data.
    fitting_axes : int or tuple
        The axes to keep for the fitting (other axes will be sliced/iterated over)
    world : `None` or tuple or APE-14-WCS
        This can be specified either as a tuple of world coordinates for each
        fitting axis, or as WCS for the whole cube. If specified as a tuple,
        the values in the tuple can be either 1D arrays, or can be given as
        N-dimensional arrays with shape broadcastable to the data shape. If
        specified as a WCS, the WCS can have any dimensionality so long as it
        matches the data. If not specified, the fitting is carried out in pixel
        coordinates.
    chunk_n_max : int
        Maximum number of fits to include in a chunk. If this is made too
        large, then the workload will not be split properly over processes, and
        if it is too small it may be inefficient. If not specified, this will
        default to 500.
    diagnostics : { None | 'error' | 'error+warn' | 'all' }, optional
        Whether to output diagnostic information for fits. This can be either
        `None` (nothing), ``'error'`` (output information for fits that raised
        exceptions), or ``'all'`` (output information for all fits).
    diagnostics_path : str, optional
        If ``diagnostics`` is not `None`, this should be the path to a folder in
        which a folder will be made for each fit that is output.
    diagnostics_callable : callable
        By default, any warnings or errors are output to ``diagnostics_path``.
        However, you can also specify a callable that can e.g. make a plot or
        write out information in a custom format. The callable should take the
        following arguments: the path to the subfolder of ``diagnostics_path``
        for the specific index being fit, a list of the coordinates passed to
        the fitter, the data array, the weights array (or `None` if no weights
        are being used), the model that was fit (or `None` if the fit errored),
        and a dictionary of other keyword arguments passed to the fitter.
    scheduler : str, optional
        If not specified, a local multi-processing scheduler will be
        used. If ``'default'``, whatever is the current default scheduler will be
        used. You can also set this to anything that would be passed to
        ``array.compute(scheduler=...)``
    fitter_kwargs : None or dict
        Keyword arguments to pass to the fitting when it is called.
    preserve_native_chunks : bool, optional
        If `True`, the native data chunks will be used, although an error will
        be raised if this chunk size does not include the whole fitting axes.
    equivalencies : list of tuple
        Any equivalencies to take into account in unit conversions
    fit_info : bool or str or iterable, optional
        Option to control whether fit information set on the ``.fit_info``
        attribute of the fitters for individual fits should be concatenated
        and set on the ``.fit_info`` on the input fitter object. The options
        are as follows:

        * `False`: don't set ``.fit_info`` on the fitter
        * `True`: set ``.fit_info`` on the fitter with all available information
           from individual fits
        * An iterable of strings: only save the specific attributes mentioned
          in the iterable

        If not `False`, the ``.fit_info`` attribute on the fitter will be set
        to a `FitInfoArrayContainer` object which can be used to query the
        fit information for individual fits. Otherwise, ``.fit_info`` will be
        left unchanged.
    """
    try:
        import dask
        import dask.array as da
    except ImportError:  # pragma: no cover
        raise ImportError("dask is required for this function")

    if scheduler is None:
        scheduler = "processes"

    if diagnostics in (None, "error", "error+warn", "all"):
        if diagnostics is not None:
            if diagnostics_path is None:
                raise ValueError("diagnostics_path should be set")
            else:
                Path(diagnostics_path).mkdir(parents=True, exist_ok=True)
    else:
        raise ValueError("diagnostics should be None, 'error', 'error+warn', or 'all'")

    original_fit_info = deepcopy(fitter.fit_info)

    if not isinstance(fitting_axes, tuple):
        fitting_axes = (fitting_axes,)

    # Check dimensionality
    if model.n_inputs != len(fitting_axes):
        raise ValueError(
            f"Model is {model.n_inputs}-dimensional, but got "
            f"{len(fitting_axes)} value(s) in fitting_axes="
        )

    for fi in fitting_axes:
        if fi <= -data.ndim or fi > data.ndim - 1:
            raise ValueError(
                f"Fitting index {fi} out of range for {data.ndim}-dimensional data"
            )

    if preserve_native_chunks:
        if not isinstance(data, da.core.Array):
            raise TypeError(
                "Can only set preserve_native_chunks=True if input data is a dask array"
            )
        if weights is not None and not isinstance(weights, da.core.Array):
            raise TypeError(
                "Can only set preserve_native_chunks=True if input weights is a dask array (if specified)"
            )

    if isinstance(weights, NDUncertainty):
        weights = weights.represent_as(StdDevUncertainty)
        weights = 1 / weights.array

    if mask is not None:
        imask = np.logical_not(mask).astype(float)
        if weights is None:
            weights = imask
        else:
            weights *= imask

    # Sanitize fitting_axes and determine iterating_axes
    ndim = data.ndim
    fitting_axes = tuple([(fi if fi >= 0 else ndim - fi) for fi in fitting_axes])
    iterating_axes = tuple([i for i in range(ndim) if i not in fitting_axes])

    # Determine the shape along the fitting dimensions and the iterating dimensions
    fitting_shape = tuple([data.shape[i] for i in fitting_axes])
    iterating_shape = tuple([data.shape[i] for i in iterating_axes])

    if data_unit is None and isinstance(data, u.Quantity):
        data_unit = data.unit
        data = data.value

    if preserve_native_chunks:
        for idx in fitting_axes:
            if data.chunksize[idx] != data.shape[idx]:
                raise ValueError(
                    "When using preserve_native_chunks=True, the chunk size should match the data size along the fitting axes"
                )
        if weights is not None and data.chunksize != weights.chunksize:
            raise ValueError(
                "When using preserve_native_chunks=True, the weights should have the same chunk size as the data"
            )
    else:
        # Rechunk the array so that it is not chunked along the fitting axes
        chunk_shape = tuple(
            "auto" if idx in iterating_axes else -1 for idx in range(ndim)
        )

        if chunk_n_max is None:
            chunk_n_max = 500

        block_size_limit = chunk_n_max * prod(fitting_shape) * data.dtype.itemsize

        if isinstance(data, da.core.Array):
            data = data.rechunk(chunk_shape, block_size_limit=block_size_limit)
        else:
            with dask.config.set({"array.chunk-size": block_size_limit}):
                data = da.from_array(data, chunks=chunk_shape, name="data")

        if weights is not None:
            if isinstance(weights, da.core.Array):
                weights = weights.rechunk(data.chunksize)
            else:
                weights = da.from_array(weights, chunks=data.chunksize, name="weights")

    world_arrays = False
    if isinstance(world, BaseHighLevelWCS):
        world = world.low_level_wcs

    if isinstance(world, BaseLowLevelWCS):
        if world.pixel_n_dim != data.ndim:
            raise ValueError(
                f"The WCS pixel_n_dim ({world.pixel_n_dim}) does not match the number of dimensions in the data ({data.ndim})"
            )

        # Note that in future we could in principle consider supporting cases
        # where the number of world dimensions does not match the number of
        # data dimensions, provided the model returns a different number of
        # outputs than it takes inputs. For example, one could consider fitting
        # a 1D dataset with a model that takes two inputs and returns one
        # output if the WCS provides two coordinates for each 1D pixel.
        # However, this is a very advanced and unusual use case, so we don't
        # cater for this for now.

        fitting_world = SlicedLowLevelWCS(
            world,
            [slice(None) if i in fitting_axes else 0 for i in range(world.pixel_n_dim)],
        )
        if fitting_world.world_n_dim != len(fitting_axes):
            raise ValueError(
                "The number of WCS world axes corresponding to the fitting axes "
                f"({fitting_world.world_n_dim}) does not match the number of fitting axes ({len(fitting_axes)})"
            )

        world_units = list(map(u.Unit, fitting_world.world_axis_units[::-1]))

        # Construct dask arrays of world coordinates for every pixel in the cube.
        # We will then iterate over this in map_blocks.
        # NOTE: This returns in world (cartesian) order
        world_dask_arrays = _wcs_to_world_dask(world, data.shape, chunks=data.chunksize)

        # Extract world arrays for just fitting dimensions
        fitting_pixel_axes = np.arange(data.ndim)[::-1][np.array(fitting_axes)]
        world_idx = [
            np.argwhere(world.axis_correlation_matrix[:, fpa])[:, 0][0]
            for fpa in fitting_pixel_axes
        ]
        world = [world_dask_arrays[idx] for idx in world_idx]
        world_arrays = True

    elif isinstance(world, tuple):
        # If world is a tuple then we allow N inputs where N is the number of fitting_axes
        # Each array in the tuple should with be broadcastable to the shape of the fitting_axes
        # or it should be one dimensional and the broadcasting can happen later
        if len(world) != len(fitting_axes):
            raise ValueError(
                f"The number of world arrays ({len(world)}) must match "
                f"number of fitting axes ({len(fitting_axes)})"
            )
        world = list(world)
        world_units = []
        for iw, w in enumerate(world):
            if (unit := getattr(w, "unit", None)) is not None:
                world[iw] = w.value
            world_units.append(unit)

        if all(w.ndim == 1 for w in world):
            for i, (w, fit_shape) in zip(fitting_axes, zip(world, fitting_shape)):
                if w.shape[0] != fit_shape:
                    raise ValueError(
                        f"world[{i}] has length {w.shape[0]} but data along "
                        f"dimension {i} has length {fit_shape}"
                    )
            world_arrays = False
        else:
            for w in world:
                try:
                    w = np.broadcast_shapes(w.shape, data.shape)
                except ValueError as e:
                    raise ValueError(
                        f"The arrays in the world tuple should be broadcastable to "
                        f"the shape of the data (expected {data.shape}), got {w.shape})"
                    ) from e
            # Extract world arrays for just fitting dimensions
            world = [da.asarray(world[idx]) for idx in fitting_axes]
            world_arrays = True
    elif world is None:
        world = tuple([np.arange(size) for size in fitting_shape])
        world_units = [None] * len(fitting_axes)
    else:
        raise TypeError("world should be None, a WCS object or a tuple of arrays")

    if model._has_units or data_unit is not None:
        # We now combine any instance-level input equivalencies with user
        # specified ones at call-time.

        input_units_equivalencies = _combine_equivalency_dict(
            model.inputs, equivalencies, model.input_units_equivalencies
        )

        # If input_units is defined, we transform the input data into those
        # expected by the model. We hard-code the input names 'x', and 'y'
        # here since FittableModel instances have input names ('x',) or
        # ('x', 'y')

        if model.input_units is None:
            target_units = world_units[:]
        else:
            target_units = [
                model.input_units[model.inputs[i]] for i in range(model.n_inputs)
            ]
            world = [
                unit.to(
                    target_units[i],
                    equivalencies=input_units_equivalencies[model.inputs[i]],
                    value=w,
                )
                if unit is not None
                else w
                for i, (w, unit) in enumerate(zip(world, world_units))
            ]

        # Create a dictionary mapping the real model inputs and outputs
        # names to the data. This remapping of names must be done here, after
        # the input data is converted to the correct units.
        rename_data = {}
        rename_data[model.inputs[0]] = (0,) * target_units[0]
        rename_data[model.outputs[0]] = (0,) * data_unit
        if len(world) == 2:
            rename_data[model.inputs[1]] = (0,) * target_units[1]
        else:
            rename_data["z"] = None

        # We now strip away the units from the parameters, taking care to
        # first convert any parameters to the units that correspond to the
        # input units (to make sure that initial guesses on the parameters)
        # are in the right unit system
        model = model.without_units_for_data(**rename_data)
        add_back_units = True

    else:
        world_units = tuple(None for w in world)
        add_back_units = False

    # Extract the parameters arrays from the model, in the order in which they
    # appear in param_names. We need to broadcast these up to the data shape so
    # that map_blocks can then iterate over both the data and parameters. We
    # need to rechunk to the same chunk shape as the data so that chunks line
    # up and for map_blocks to work properly as noted in
    # https://github.com/dask/dask/issues/11188. However, rather than use dask
    # operations to broadcast these up, which creates a complex graph and
    # results in high memory usage, we use a ParameterContainer which does the
    # broadcasting on-the-fly as needed.
    parameter_arrays = []
    for name in model.param_names:
        values = getattr(model, name).value
        parameter_arrays.append(
            da.from_array(
                ParameterContainer(values, iterating_shape, iterating_axes, data.shape),
                chunks=data.chunksize,
                name="parameter-" + name,
            )
        )

    # Define a model with default parameters to pass in to _fit_models_to_chunk without copying all the parameter data

    simple_model = _copy_with_new_parameters(model, {})

    weights_array = [] if weights is None else [weights]

    result = da.map_blocks(
        _fit_models_to_chunk,
        data,
        *weights_array,
        *parameter_arrays,
        *(world if world_arrays else []),
        enforce_ndim=True,
        dtype=object if fit_info else float,
        drop_axis=fitting_axes,
        model=simple_model,
        fitter=fitter,
        new_axis=0,
        world=world if not world_arrays else None,
        diagnostics=diagnostics,
        diagnostics_path=diagnostics_path,
        diagnostics_callable=diagnostics_callable,
        iterating_shape=iterating_shape,
        iterating_axes=iterating_axes,
        fitting_axes=fitting_axes,
        fitter_kwargs=fitter_kwargs,
        name="fitting-results",
        weights_specified=weights is not None,
        fit_info=fit_info,
    )

    if scheduler == "default":
        compute_kwargs = {}
    else:
        compute_kwargs = {"scheduler": scheduler}

    result_array = result.compute(**compute_kwargs)

    if fit_info:
        parameter_arrays_fitted = result_array[:-1].astype(float)
        fit_info_array = result_array[-1]
    else:
        parameter_arrays_fitted = result_array

    # Set up new parameter arrays with fitted values
    parameters = {}
    for i, name in enumerate(model.param_names):
        parameters[name] = parameter_arrays_fitted[i].reshape(iterating_shape)

    if fit_info:
        fitter.fit_info = FitInfoArrayContainer(fit_info_array)
    else:
        fitter.fit_info = original_fit_info

    # Instantiate new fitted model
    model_fitted = _copy_with_new_parameters(model, parameters)

    # Add back units if needed
    if add_back_units:
        model_fitted = model_fitted.with_units_from_data(**rename_data)

    return model_fitted



from pathlib import Path

import numpy as np
import pytest

from astropy import units
from astropy.nddata import covariance
from astropy.table import Table
from astropy.utils.compat.optional_deps import HAS_SCIPY
from astropy.utils.exceptions import AstropyUserWarning

scipy_required = pytest.mark.skipif(not HAS_SCIPY, reason="scipy not installed")


def mock_cov():
    # Build a bogus covariance matrix
    return (
        np.diag(np.full(10 - 2, 0.2, dtype=float), k=-2)
        + np.diag(np.full(10 - 1, 0.5, dtype=float), k=-1)
        + np.diag(np.full(10, 1.0, dtype=float), k=0)
        + np.diag(np.full(10 - 1, 0.5, dtype=float), k=1)
        + np.diag(np.full(10 - 2, 0.2, dtype=float), k=2)
    )


def mock_cov_2d():
    # Build a bogus covariance matrix that is for a 3,2 2D array
    c = (
        np.diag(np.full(6 - 2, 0.2, dtype=float), k=-2)
        + np.diag(np.full(6 - 1, 0.5, dtype=float), k=-1)
        + np.diag(np.full(6, 1.0, dtype=float), k=0)
        + np.diag(np.full(6 - 1, 0.5, dtype=float), k=1)
        + np.diag(np.full(6 - 2, 0.2, dtype=float), k=2)
    )
    return (3, 2), c


def mock_cov_3d():
    # Build a bogus covariance matrix that is for a 3,2,3 3D array
    c = (
        np.diag(np.full(18 - 2, 0.2, dtype=float), k=-2)
        + np.diag(np.full(18 - 1, 0.5, dtype=float), k=-1)
        + np.diag(np.full(18, 1.0, dtype=float), k=0)
        + np.diag(np.full(18 - 1, 0.5, dtype=float), k=1)
        + np.diag(np.full(18 - 2, 0.2, dtype=float), k=2)
    )
    return (3, 2, 3), c


def test_scipy_funcs():
    # Functions should always be callable, regardless of whether or not scipy is installed
    assert callable(covariance.find), "find should be callable"
    assert callable(covariance.triu), "triu should be callable"
    assert callable(covariance.csr_matrix), "csr_matrix should be callable"
    assert callable(covariance.coo_matrix), "coo_matrix should be callable"
    assert callable(covariance.isspmatrix_csr), "isspmatrix_csr should be callable"

    if not HAS_SCIPY:
        # Should raise an error if scipy is not available
        with pytest.raises(ModuleNotFoundError):
            covariance.find()
        with pytest.raises(ModuleNotFoundError):
            covariance.triu()
        with pytest.raises(ModuleNotFoundError):
            covariance.csr_matrix()
        with pytest.raises(ModuleNotFoundError):
            covariance.coo_matrix()
        with pytest.raises(ModuleNotFoundError):
            covariance.isspmatrix_csr()


@scipy_required
def test_get_csr():
    c = mock_cov()
    # Convert a numpy array
    carr = covariance._get_csr(c)
    assert np.array_equal(carr.toarray(), c), "np.array not converted properly"
    # Convert a list
    clist = covariance._get_csr(c.tolist())
    assert np.array_equal(clist.toarray(), c), "List not converted properly"
    # Return input array if already a csr_matrix
    ccsr = covariance.csr_matrix(c)
    _ccsr = covariance._get_csr(ccsr)
    assert _ccsr is ccsr, "Should return reference to the input"
    # Fail if cannot be converted
    with pytest.raises(TypeError):
        cov = covariance.Covariance(array="test")


@scipy_required
def test_impose_sparse_value_threshold():
    c = covariance.csr_matrix(mock_cov())

    _c = covariance._impose_sparse_value_threshold(c, 0.0)
    assert c is _c, "All elements are above 0, so array should be returned exactly."

    _c = covariance._impose_sparse_value_threshold(c, 0.21)
    assert c is not _c, "Should remove elements"
    assert _c.nnz + 16 == c.nnz, "Incorrect number of elements removed."

    _c = covariance._impose_sparse_value_threshold(c, 0.51)
    assert c is not _c, "Should remove elements"
    assert _c.nnz == _c.shape[0], "Should remove all but the diagonal"


def test_parse_shape():
    shape = "(2,1)"
    assert covariance._parse_shape(shape) == (2, 1), f"Bad parse for {shape}"
    shape = "(2,1,5)"
    assert covariance._parse_shape(shape) == (2, 1, 5), f"Bad parse for {shape}"
    shape = "(2,1,5,)"
    assert covariance._parse_shape(shape) == (2, 1, 5), f"Bad parse for {shape}"
    shape = "(2,  1,  5,)"
    assert covariance._parse_shape(shape) == (2, 1, 5), f"Bad parse for {shape}"


@scipy_required
def test_ingest():
    # It must be possible to convert the input array into a csr_matrix
    with pytest.raises(TypeError):
        cov = covariance.Covariance._ingest_matrix("test")

    # Must be 2D.  If 1D, the array is successfully converted to a csr_matrix,
    # but fails the 2D check.  So the error is ValueError.
    arr1d = np.ones(27, dtype=float)
    with pytest.raises(ValueError):
        cov = covariance.Covariance._ingest_matrix(arr1d)

    # Must be 2D.  If 3D (or higher), the array fails to be converted to a
    # csr_matrix, so the error is TypeError.
    arr3d = np.ones(27, dtype=float).reshape(3, 3, 3)
    with pytest.raises(TypeError):
        cov = covariance.Covariance._ingest_matrix(arr3d)

    # Must be square
    with pytest.raises(ValueError):
        cov = covariance.Covariance._ingest_matrix(arr3d.reshape(9, 3))

    # Should be symmetric, but just throw a warning for now
    c = np.array([[1, 0], [1, 1]]).astype(float)
    with pytest.warns(AstropyUserWarning):
        cov = covariance.Covariance._ingest_matrix(c)

    # Can skip the symmetry check, the warning won't be thrown
    cov = covariance.Covariance._ingest_matrix(c, assume_symmetric=True)
    assert np.array_equal(cov.toarray(), c), (
        "_ingest_matrix should not correct symmetry issue if the check is skipped"
    )


@scipy_required
def test_bad_init_type():
    # Cannot be None
    with pytest.raises(ValueError):
        cov = covariance.Covariance()

    # Should be symmetric, but just throw a warning for now
    c = np.array([[1, 0], [1, 1]]).astype(float)
    with pytest.warns(AstropyUserWarning):
        cov = covariance.Covariance(array=c)

    # If the symmetry check is skipped, the warning won't be thrown but the
    # symmetry will be imposed anyway.
    cov = covariance.Covariance(array=c, assume_symmetric=True)
    assert np.array_equal(cov.to_dense(), np.identity(2)), (
        "Instantiation should correct symmetry"
    )


@scipy_required
def test_shape_mismatch():
    data_shape, c = mock_cov_2d()
    bad_shape = (2, 2)
    assert bad_shape != data_shape, (
        "Shapes should not match for the test to work properly"
    )
    with pytest.raises(ValueError):
        cov = covariance.Covariance(array=c, data_shape=bad_shape)


@scipy_required
def test_uncertainty_string():
    cov = covariance.Covariance(array=mock_cov())
    assert cov.uncertainty_type == "cov", "Uncertainty type changed"


@scipy_required
def test_quantity():
    data_shape, c = mock_cov_2d()
    cov = covariance.Covariance(array=c, data_shape=data_shape, unit="Jy^2")
    q = cov.quantity

    assert isinstance(q, units.Quantity), "Wrong type"
    assert q.unit == cov.unit, "Unit mismatch"
    assert np.array_equal(q.value, cov.to_dense()), "Array mismatch"


@scipy_required
def test_init():
    c = mock_cov()

    # Should not fault if directly instantiated from a numpy.array
    cov = covariance.Covariance(array=c)

    # Convert to a CSR sparse matrix
    c_csr = covariance.csr_matrix(c)
    # And instantiate
    cov = covariance.Covariance(array=c_csr)

    # Directly access the covariance array, convert it to a dense matrix, and
    # check it against the original array.
    assert np.array_equal(cov.to_dense(), c), "Ingested array does not match input"


@scipy_required
def test_stored_nnz():
    c = mock_cov()
    cov = covariance.Covariance(array=c)
    assert cov.stored_nnz == np.sum(np.triu(c) > 0), (
        "Number of stored non-zero elements differ"
    )


@scipy_required
def test_nnz():
    c_csr = covariance.csr_matrix(mock_cov())
    cov = covariance.Covariance(array=c_csr)
    assert c_csr.nnz == cov.nnz, "Number of non-zero elements differ"


@scipy_required
def test_indices():
    # Test when data_shape is not defined
    cov = covariance.Covariance(array=mock_cov())

    # Test out of bounds indices
    i = np.array([10, 1, 2])
    j = np.array([13, 4, 3])
    with pytest.raises(ValueError):
        cov.covariance_to_data_indices(i, j)
    with pytest.raises(ValueError):
        cov.data_to_covariance_indices(i, j)

    # Test in bounds
    i_cov = np.array([0, 1, 2])
    j_cov = np.array([3, 4, 3])
    i_data, j_data = cov.covariance_to_data_indices(i_cov, j_cov)
    assert np.array_equal(i_cov, i_data) and np.array_equal(j_cov, j_data), (
        "Should return output"
    )
    _i_cov, _j_cov = cov.data_to_covariance_indices(i_data, j_data)
    assert np.array_equal(_i_cov, i_data) and np.array_equal(_j_cov, j_data), (
        "Should return output"
    )

    # Test multi-dimensional data
    data_shape, c = mock_cov_2d()
    cov = covariance.Covariance(array=c, data_shape=data_shape)
    i_cov = np.array([0, 1, 2])
    j_cov = np.array([3, 4, 3])
    i_data, j_data = cov.covariance_to_data_indices(i_cov, j_cov)
    assert len(i_data) == 2, "Should return indices for each of 2 dimensions."
    assert (i_data[0][0], i_data[1][0]) == (0, 0), "Coordinate conversion error"

    _i_cov, _j_cov = cov.data_to_covariance_indices(i_data, j_data)
    assert np.array_equal(i_cov, _i_cov), "Inverse operation error"
    assert np.array_equal(j_cov, _j_cov), "Inverse operation error"

    # Test shape mismatch
    with pytest.raises(ValueError):
        cov.data_to_covariance_indices(i_data + (np.array([0, 0, 0]),), j_data)
    with pytest.raises(ValueError):
        cov.data_to_covariance_indices(i_data, j_data + (np.array([0, 0, 0]),))


@scipy_required
def test_coo():
    # 1D
    cov = covariance.Covariance(array=covariance.csr_matrix(mock_cov()))
    i, j, cij = cov.coordinate_data()
    assert i.size == cov.stored_nnz, "Coordinate data length is the incorrect size"

    # Cannot reshape when data_shape is not defined
    with pytest.raises(ValueError):
        cov.coordinate_data(reshape=True)

    # 2D
    data_shape, c = mock_cov_2d()
    c_csr = covariance.csr_matrix(c)
    cov = covariance.Covariance(array=c_csr, data_shape=data_shape)
    # Try without reshaping
    ic, jc, cij = cov.coordinate_data(reshape=False)
    assert isinstance(ic, np.ndarray), (
        "Index object should be an array if not reshaping"
    )
    assert ic.size == cov.stored_nnz, "Incorrect number of non-zero elements"
    # Try with reshaping
    i, j, cij = cov.coordinate_data(reshape=True)
    assert len(i) == len(data_shape), "Dimensionality does not match"
    assert i[0].size == cov.stored_nnz, "Incorrect number of non-zero elements"

    # Make sure we recover the same covariance matrix indices
    assert np.array_equal(ic, np.ravel_multi_index(i, cov.data_shape)), (
        "Bad covariance index mapping"
    )

    # 3D
    data_shape, c = mock_cov_3d()
    c_csr = covariance.csr_matrix(c)
    cov = covariance.Covariance(array=c_csr, data_shape=data_shape)
    i, j, cij = cov.coordinate_data(reshape=True)
    assert len(i) == len(data_shape), "Dimensionality does not match"
    assert i[0].size == cov.stored_nnz, "Incorrect number of non-zero elements"


@scipy_required
def test_copy():
    cov = covariance.Covariance(array=covariance.csr_matrix(mock_cov()))
    _cov = cov.copy()
    assert cov is not _cov, "Objects have the same reference"
    assert cov._cov is not _cov._cov, "Object arrays have the same reference"
    assert np.array_equal(cov.to_dense(), _cov.to_dense()), "Arrays should be equal"


@scipy_required
def test_tbls():
    cov = covariance.Covariance(array=covariance.csr_matrix(mock_cov()))
    covar = cov.to_table()
    assert isinstance(covar, Table), "correlation data should be output as a table"
    assert len(covar) == np.sum(np.triu(mock_cov()) > 0), (
        "Incorrect number of table entries"
    )
    assert len(covar.colnames) == 3, "Incorrect number of columns"
    assert covar["INDXI"].ndim == 1, "Incorrect shape for index array"

    _cov = covariance.Covariance.from_table(covar)
    assert np.array_equal(cov.to_dense(), _cov.to_dense()), (
        "Bad convert/revert from tables"
    )

    # Test failure when meta is None
    with pytest.raises(ValueError):
        _covar = covar.copy()
        _covar.meta = {}
        _cov = covariance.Covariance.from_table(_covar)

    data_shape, c = mock_cov_3d()
    cov = covariance.Covariance(array=covariance.csr_matrix(c), data_shape=data_shape)
    covar = cov.to_table()
    assert len(covar) == np.sum(np.triu(c) > 0), "Incorrect number of table entries"
    assert len(covar.colnames) == 3, "Incorrect number of columns"
    assert covar["INDXI"].ndim == 2, "Incorrect shape for index array"

    _cov = covariance.Covariance.from_table(covar)
    assert np.array_equal(cov.to_dense(), _cov.to_dense()), (
        "Bad convert/revert from tables"
    )

    # Introduce a shape mismatch
    covar.meta["COVDSHP"] = "(3,6)"
    with pytest.raises(ValueError):
        _cov = covariance.Covariance.from_table(covar)


@scipy_required
def test_samples():
    # Fixed-seed RNG for repeatability
    rng = np.random.default_rng(seed=8001)

    m = np.zeros(10, dtype=float)
    c = mock_cov()

    # Shape must be 2D
    s = rng.multivariate_normal(m, c)
    with pytest.raises(ValueError):
        covar = covariance.Covariance.from_samples(s, cov_tol=0.1)

    # Shape must be at least two samples
    s = np.expand_dims(s, 1).T
    with pytest.raises(ValueError):
        covar = covariance.Covariance.from_samples(s.T, cov_tol=0.1)

    # Draw samples
    s = rng.multivariate_normal(m, c, size=100000)

    # Instantiate
    covar = covariance.Covariance.from_samples(s.T, cov_tol=0.1)

    # Check the values are very nearly the same as the input
    assert np.all(np.absolute(c - covar.to_dense()) < 0.02), (
        "Covariances are too different"
    )

    # Check that `find` returns the same indices
    # NOTE: For some reason the ordering of np.where and
    # scipy.sparse.find are different.  So I run lexsort before checking
    # if the arrays are the same.

    coo = np.array([[i, j] for i, j in zip(*np.where(c > 0))])
    srt = np.lexsort((coo[:, 0], coo[:, 1]))
    coo = coo[srt, :]

    _coo = np.array([[i, j] for i, j, k in zip(*covar.find())])
    srt = np.lexsort((_coo[:, 0], _coo[:, 1]))
    _coo = _coo[srt, :]

    assert np.array_equal(coo, _coo), "Did not find the same indices"


@scipy_required
def test_mult():
    c = mock_cov()
    x = np.ones(10, dtype=float)

    # Transfer matrix must be 2D
    t = np.ones(3, dtype=float)
    with pytest.raises(ValueError):
        covar = covariance.Covariance.from_matrix_multiplication(t, c)

    # Uncorrelated
    t = np.zeros((3, 10), dtype=float)
    t[0, 0] = 1.0
    t[1, 3] = 1.0
    t[2, 6] = 1.0

    # Sigma has the wrong shape
    with pytest.raises(ValueError):
        covar = covariance.Covariance.from_matrix_multiplication(t, c[1:, 1:])
    with pytest.raises(ValueError):
        covar = covariance.Covariance.from_matrix_multiplication(t, np.diag(c)[1:])

    y = np.dot(t, x)

    covar = covariance.Covariance.from_matrix_multiplication(t, c)
    assert np.array_equal(covar.to_dense(), np.identity(3)), (
        "Result should be uncorrelated."
    )

    # Correlated by 0.2
    t = np.zeros((3, 10), dtype=float)
    t[0, 0] = 1.0
    t[1, 2] = 1.0
    t[2, 4] = 1.0
    _c = (
        np.diag(np.full(3 - 1, 0.2, dtype=float), k=-1)
        + np.diag(np.full(3, 1.0, dtype=float), k=0)
        + np.diag(np.full(3 - 1, 0.2, dtype=float), k=1)
    )
    y = np.dot(t, x)
    covar = covariance.Covariance.from_matrix_multiplication(t, c)
    assert np.array_equal(covar.to_dense(), _c), (
        "Result should have off-diagonals = 0.2"
    )

    # Correlated by 0.5 and 0.2
    t = np.zeros((3, 10), dtype=float)
    t[0, 0] = 1.0
    t[1, 1] = 1.0
    t[2, 2] = 1.0
    _c = (
        np.diag(np.full(3 - 2, 0.2, dtype=float), k=-2)
        + np.diag(np.full(3 - 1, 0.5, dtype=float), k=-1)
        + np.diag(np.full(3, 1.0, dtype=float), k=0)
        + np.diag(np.full(3 - 1, 0.5, dtype=float), k=1)
        + np.diag(np.full(3 - 2, 0.2, dtype=float), k=2)
    )
    y = np.dot(t, x)
    covar = covariance.Covariance.from_matrix_multiplication(t, c)
    assert np.array_equal(covar.to_dense(), _c), (
        "Result should have off-diagonals = 0.5,0.2"
    )


@scipy_required
def test_var():
    var = np.ones(3, dtype=float)
    covar = covariance.Covariance.from_variance(var)
    assert np.array_equal(covar.to_dense(), np.identity(3)), (
        "Result should be an identity matrix"
    )


@scipy_required
def test_shape():
    cov = covariance.Covariance(array=mock_cov())
    assert len(cov.data_shape) == 1, "Incorrect dimensionality"
    assert cov.data_shape == (mock_cov().shape[0],), "Bad data shape"

    # 2D
    data_shape, c = mock_cov_2d()
    cov = covariance.Covariance(array=c, data_shape=data_shape)
    assert len(cov.data_shape) == 2, "Incorrect dimensionality"
    assert cov.data_shape == data_shape, "Bad data shape"

    # 3D
    data_shape, c = mock_cov_3d()
    cov = covariance.Covariance(array=c, data_shape=data_shape)
    assert len(cov.data_shape) == 3, "Incorrect dimensionality"
    assert cov.data_shape == data_shape, "Bad data shape"


@scipy_required
def test_match_to_data_slice():
    c = mock_cov()
    cov = covariance.Covariance(array=c)

    # Check the index map
    assert np.array_equal(cov.data_index_map, np.arange(10)), "Bad index map"

    # 1D slice
    sub_cov = cov.match_to_data_slice(np.s_[:5])
    assert isinstance(sub_cov, covariance.Covariance), (
        "Submatrix should be a Covariance instance"
    )
    assert np.array_equal(sub_cov.to_dense(), c[:5, :5]), "Bad submatrix"

    # 1D reorder
    rng = np.random.default_rng(99)
    reorder = np.arange(cov.shape[0])
    rng.shuffle(reorder)
    sub_cov = cov.match_to_data_slice(reorder)
    assert isinstance(sub_cov, covariance.Covariance), (
        "Submatrix should be a Covariance instance"
    )
    assert np.array_equal(sub_cov.to_dense(), c[np.ix_(reorder, reorder)]), (
        "Bad matrix reorder"
    )

    # 2D
    data_shape, c = mock_cov_2d()
    cov = covariance.Covariance(array=c, data_shape=data_shape)

    # Check the index map
    assert cov.data_index_map.shape == data_shape, "Bad index map shape"
    assert np.array_equal(cov.data_index_map, np.arange(6).reshape(data_shape)), (
        "Bad index map"
    )

    # Reduce dimensionality
    sub_cov = cov.match_to_data_slice(np.s_[:, 0])
    assert len(sub_cov.data_shape) == 1, "Submatrix should have reduced dimensionality"
    assert sub_cov.shape == (3, 3), "Incorrect covariance submatrix shape"
    # Still 2D
    sub_cov = cov.match_to_data_slice(np.s_[1:, :])
    assert sub_cov.data_shape == (2, 2), "Submatrix does not have the correct shape"

    # 3D
    data_shape, c = mock_cov_3d()
    cov = covariance.Covariance(array=c, data_shape=data_shape)
    # Reduce dimensionality to 1D
    sub_cov = cov.match_to_data_slice(np.s_[:, 0, 0])
    assert len(sub_cov.data_shape) == 1, "Submatrix should have reduced dimensionality"
    assert sub_cov.shape == (3, 3), "Incorrect covariance submatrix shape"
    # Reduce dimensionality to 2D
    sub_cov = cov.match_to_data_slice(np.s_[1:, :, 0])
    assert sub_cov.data_shape == (2, 2), "Submatrix does not have the correct shape"
    # Still 3D
    sub_cov = cov.match_to_data_slice(np.s_[1:, :, :-1])
    assert sub_cov.data_shape == (2, 2, 2), "Submatrix does not have the correct shape"
    assert sub_cov.shape[0] == np.prod(sub_cov.data_shape), "Shape mismatch"


@scipy_required
def test_correl():
    # Since the diagonal is 1, this is actually a correlation matrix
    c = mock_cov()
    # Set two covariance arrays with different variances but the same correlations
    cov1 = covariance.Covariance(array=c * 4.0)
    cov2 = covariance.Covariance(array=c * 2.0)
    rho1 = cov1.to_sparse(correlation=True)
    # Should be the same as the input
    assert np.allclose(rho1.toarray(), c), "Correlation matrix changed"

    # Should match cov1
    rho2 = cov2.to_sparse(correlation=True)
    assert np.allclose(rho1.toarray(), rho2.toarray()), (
        "Correlation matrices should be identical"
    )
    assert np.allclose(cov1.variance / 2, cov2.variance), "Variances incorrect"


@scipy_required
def test_to_from_correl():
    c = mock_cov()
    cov1 = covariance.Covariance(array=c * 4.0)

    var, rho1 = covariance.Covariance.to_correlation(cov1.to_sparse())
    assert np.array_equal(rho1.toarray(), c), "Bad correlation matrix recovery"

    _cov1 = covariance.Covariance.revert_correlation(var, rho1)
    assert np.array_equal(cov1.to_dense(), _cov1.toarray())


@scipy_required
def test_newvar():
    c = mock_cov()
    cov1 = covariance.Covariance(array=c)
    var = np.full(c.shape[0], 4.0, dtype=float)
    with pytest.raises(NotImplementedError):
        cov1.variance = var
    cov2 = cov1.apply_new_variance(var)
    assert np.allclose(var, cov2.variance), "Variance does not match request"
    var2, rho2 = covariance.Covariance.to_correlation(cov2.to_dense())
    assert np.allclose(cov1.to_dense(), rho2.toarray()), (
        "Correlation matrices do not match"
    )


@scipy_required
def test_array():
    # Construct the Covariance matrix from a pre-calculated array
    c = mock_cov()
    covar = covariance.Covariance.from_array(c)
    # Should be the same as the identity matrix.
    assert np.array_equal(covar.to_dense(), c), "Arrays should be identical"

    # Copy
    _c = c.copy()
    # Introduce an asymmetry in the upper triangle
    _c[0, 9] = 1.0
    with pytest.warns(AstropyUserWarning):
        cov = covariance.Covariance.from_array(_c)

    # Instantiate again but assume symmetry
    cov = covariance.Covariance.from_array(_c, assume_symmetric=True)
    assert not np.array_equal(cov.to_dense(), c), (
        "Asymmetries in the upper triangle should be kept"
    )
    _c = c.copy()
    # Introduce an asymmetry in the lower triangle
    _c[9, 0] = 1.0
    cov = covariance.Covariance.from_array(_c, assume_symmetric=True)
    assert np.array_equal(cov.to_dense(), c), (
        "Asymmetries in the lower triangle should be ignored"
    )

    # Test rho tolerance (cov tolerance is test elsewhere)
    rho_tol = 0.3
    covar = covariance.Covariance.from_array(c, rho_tol=rho_tol)
    assert not np.array_equal(covar.to_dense(), c), (
        "Tolerance should have removed values"
    )
    _c = covar.to_dense()
    assert not np.any((_c > 0) & (_c < rho_tol)), (
        "Array includes elements below tolerance"
    )


@scipy_required
def test_io():
    # Set the file name
    ofile = Path("test_covar_io.fits")
    # Erase it if it already exists
    if ofile.is_file():
        ofile.unlink()

    # Create the covariance object
    covar = covariance.Covariance(array=mock_cov())
    # Write it to a table
    tbl = covar.to_table()
    # Write the table to a fits file
    tbl.write(ofile, format="fits")
    # Read it back in as a Table
    _tbl = Table.read(ofile, format="fits")
    # Use the table to instantiate a Covariance object
    _covar = covariance.Covariance.from_table(_tbl)
    # Check that the IO was successful
    assert np.array_equal(covar.to_dense(), _covar.to_dense()), "Array changed"
    # Delete the file
    ofile.unlink()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pickle

import numpy as np

from astropy.time import Time


class TestPickle:
    """Basic pickle test of time"""

    def test_pickle(self):
        times = ["1999-01-01 00:00:00.123456789", "2010-01-01 00:00:00"]
        t1 = Time(times, scale="utc")

        for prot in range(pickle.HIGHEST_PROTOCOL):
            t1d = pickle.dumps(t1, prot)
            t1l = pickle.loads(t1d)
            assert np.all(t1l == t1)

        t2 = Time("2012-06-30 12:00:00", scale="utc")

        for prot in range(pickle.HIGHEST_PROTOCOL):
            t2d = pickle.dumps(t2, prot)
            t2l = pickle.loads(t2d)
            assert t2l == t2

    def test_cache_not_shared(self):
        t = Time(["2001:020", "2001:040", "2001:060", "2001:080"], out_subfmt="date")
        # Ensure something is in the cache.
        t.value
        assert "format" in t.cache
        td = pickle.dumps(t)
        assert "format" in t.cache
        tl = pickle.loads(td)
        assert "format" in t.cache
        assert "format" not in tl.cache
        t[0] = "1999:099"
        assert t.value[0] == "1999:099"
        assert tl.value[0] == "2001:020"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Astropy Cosmology. **NOT public API**.

The public API is provided by `astropy.cosmology.traits`.
"""

__all__ = [
    "BaryonComponent",
    "CriticalDensity",
    "CurvatureComponent",
    "DarkEnergyComponent",
    "DarkMatterComponent",
    "HubbleParameter",
    "MatterComponent",
    "PhotonComponent",
    "ScaleFactor",
    "TemperatureCMB",
    "TotalComponent",
]

from .baryons import BaryonComponent
from .curvature import CurvatureComponent
from .darkenergy import DarkEnergyComponent
from .darkmatter import DarkMatterComponent
from .hubble import HubbleParameter
from .matterdensity import MatterComponent
from .photoncomponent import PhotonComponent
from .rhocrit import CriticalDensity
from .scale_factor import ScaleFactor
from .tcmb import TemperatureCMB
from .totalcomponent import TotalComponent



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This package defines miscellaneous units. They are also available in
(and should be used through) the `astropy.units` namespace.
"""
# avoid ruff complaints about undefined names defined by def_unit
# ruff: noqa: F821

import numpy as np

from astropy.constants import si as _si

from . import si
from .core import binary_prefixes, def_unit, si_prefixes
from .docgen import generate_dunder_all, generate_unit_summary

__all__: list[str] = []  #  Units are added at the end

_ns = globals()

###########################################################################
# AREAS

def_unit(
    ["barn", "barn"],
    10**-28 * si.m**2,
    namespace=_ns,
    prefixes=True,
    doc="barn: unit of area used in HEP",
)


###########################################################################
# ANGULAR MEASUREMENTS

def_unit(
    ["cycle", "cy"],
    2.0 * np.pi * si.rad,
    namespace=_ns,
    prefixes=False,
    doc="cycle: angular measurement, a full turn or rotation",
)
def_unit(
    ["spat", "sp"],
    4.0 * np.pi * si.sr,
    namespace=_ns,
    prefixes=False,
    doc="spat: the solid angle of the sphere, 4pi sr",
)

##########################################################################
# PRESSURE

def_unit(
    ["bar"],
    1e5 * si.Pa,
    namespace=_ns,
    prefixes=[(["m"], ["milli"], 1.0e-3)],
    doc="bar: pressure",
)
# The torr is almost the same as mmHg but not quite.
# See https://en.wikipedia.org/wiki/Torr
def_unit(
    ["Torr", "torr"],
    _si.atm.value / 760.0 * si.Pa,
    namespace=_ns,
    prefixes=[(["m"], ["milli"], 1.0e-3)],
    doc=(
        "Unit of pressure based on an absolute scale, now defined as "
        "exactly 1/760 of a standard atmosphere"
    ),
)

###########################################################################
# MASS

def_unit(
    ["M_p"],
    _si.m_p,
    namespace=_ns,
    doc="Proton mass",
    format={"latex": r"M_{p}", "unicode": "Mₚ"},
)
def_unit(
    ["M_e"],
    _si.m_e,
    namespace=_ns,
    doc="Electron mass",
    format={"latex": r"M_{e}", "unicode": "Mₑ"},
)
def_unit(
    ["u", "Da", "Dalton"],
    _si.u,
    namespace=_ns,
    prefixes=True,
    exclude_prefixes=["a", "da"],
    doc="Unified atomic mass unit",
)

##########################################################################
# ENERGY

def_unit(
    ["eV", "electronvolt"],
    _si.e.value * si.J,
    namespace=_ns,
    prefixes=True,
    doc="Electron Volt",
)

# Here, explicitly convert the planck constant to 'eV s' since the constant
# can override that to give a more precise value that takes into account
# covariances between e and h.  Eventually, this may also be replaced with
# just `_si.Ryd.to(eV)`.
def_unit(
    ["Ry", "rydberg"],
    (_si.Ryd * _si.c * _si.h.to(eV * si.s)).to(eV),
    namespace=_ns,
    prefixes=True,
    doc="Rydberg: Energy of a photon whose wavenumber is the Rydberg constant",
    format={"latex": r"R_{\infty}", "unicode": "R∞"},
)


###########################################################################
# COMPUTER

def_unit(
    (["bit", "b"], ["bit"]),
    namespace=_ns,
    prefixes=si_prefixes + binary_prefixes,
)
def_unit(
    (["byte", "B"], ["byte"]),
    8 * bit,
    namespace=_ns,
    prefixes=si_prefixes + binary_prefixes,
    exclude_prefixes=["d"],
)
def_unit(
    (["pix", "pixel"], ["pixel"]),
    format={"ogip": "pixel", "vounit": "pixel"},
    namespace=_ns,
    prefixes=True,
)
def_unit(
    (["vox", "voxel"], ["voxel"]),
    format={"fits": "voxel", "ogip": "voxel", "vounit": "voxel"},
    namespace=_ns,
    prefixes=True,
)

###########################################################################
# ALL & DOCSTRING

__all__ += generate_dunder_all(globals())  # noqa: PLE0605


if __doc__ is not None:
    # This generates a docstring for this module that describes all of the
    # standard units defined here.
    __doc__ += generate_unit_summary(globals())



"""Various implementations of the Lomb-Scargle Periodogram."""

from .chi2_impl import lombscargle_chi2
from .fast_impl import lombscargle_fast
from .fastchi2_impl import lombscargle_fastchi2
from .main import available_methods, lombscargle
from .scipy_impl import lombscargle_scipy
from .slow_impl import lombscargle_slow



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# pylint: disable=invalid-name, no-member

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy import units as u
from astropy.modeling import models
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import wcs


@pytest.mark.parametrize(
    "inp",
    [(0, 0), (4000, -20.56), (-2001.5, 45.9), (0, 90), (0, -90), (np.mgrid[:4, :6])],
)
def test_against_wcslib(inp):
    w = wcs.WCS()
    crval = [202.4823228, 47.17511893]
    w.wcs.crval = crval
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]

    lonpole = 180
    tan = models.Pix2Sky_TAN()
    n2c = models.RotateNative2Celestial(
        crval[0] * u.deg, crval[1] * u.deg, lonpole * u.deg
    )
    c2n = models.RotateCelestial2Native(
        crval[0] * u.deg, crval[1] * u.deg, lonpole * u.deg
    )
    m = tan | n2c
    minv = c2n | tan.inverse

    radec = w.wcs_pix2world(inp[0], inp[1], 1)
    xy = w.wcs_world2pix(radec[0], radec[1], 1)

    assert_allclose(m(*inp), radec, atol=1e-12)
    assert_allclose(minv(*radec), xy, atol=1e-12)


@pytest.mark.parametrize(
    "inp", [(40 * u.deg, -0.057 * u.rad), (21.5 * u.arcsec, 45.9 * u.deg)]
)
def test_roundtrip_sky_rotation(inp):
    lon, lat, lon_pole = 42 * u.deg, (43 * u.deg).to(u.arcsec), (44 * u.deg).to(u.rad)
    n2c = models.RotateNative2Celestial(lon, lat, lon_pole)
    c2n = models.RotateCelestial2Native(lon, lat, lon_pole)
    assert_quantity_allclose(n2c.inverse(*n2c(*inp)), inp, atol=1e-13 * u.deg)
    assert_quantity_allclose(c2n.inverse(*c2n(*inp)), inp, atol=1e-13 * u.deg)


def test_Rotation2D():
    model = models.Rotation2D(angle=90 * u.deg)
    a, b = 1 * u.deg, 0 * u.deg
    x, y = model(a, b)
    assert_quantity_allclose([x, y], [0 * u.deg, 1 * u.deg], atol=1e-10 * u.deg)


def test_Rotation2D_inverse():
    model = models.Rotation2D(angle=234.23494 * u.deg)
    x, y = model.inverse(*model(1 * u.deg, 0 * u.deg))
    assert_quantity_allclose([x, y], [1 * u.deg, 0 * u.deg], atol=1e-10 * u.deg)


def test_euler_angle_rotations():
    ydeg = (90 * u.deg, 0 * u.deg)
    y = (90, 0)
    z = (0, 90)

    # rotate y into minus z
    model = models.EulerAngleRotation(0 * u.rad, np.pi / 2 * u.rad, 0 * u.rad, "zxz")
    assert_allclose(model(*z), y, atol=10**-12)
    model = models.EulerAngleRotation(0 * u.deg, 90 * u.deg, 0 * u.deg, "zxz")
    assert_quantity_allclose(model(*(z * u.deg)), ydeg, atol=10**-12 * u.deg)


@pytest.mark.parametrize(
    "params",
    [
        (60, 10, 25),
        (60 * u.deg, 10 * u.deg, 25 * u.deg),
        ((60 * u.deg).to(u.rad), (10 * u.deg).to(u.rad), (25 * u.deg).to(u.rad)),
    ],
)
def test_euler_rotations_with_units(params):
    x = 1 * u.deg
    y = 1 * u.deg
    phi, theta, psi = params

    urot = models.EulerAngleRotation(phi, theta, psi, axes_order="xyz")
    a, b = urot(x.value, y.value)
    assert_allclose((a, b), (-23.614457631192547, 9.631254579686113))
    a, b = urot(x, y)
    assert_quantity_allclose(
        (a, b), (-23.614457631192547 * u.deg, 9.631254579686113 * u.deg)
    )
    a, b = urot(x.to(u.rad), y.to(u.rad))
    assert_quantity_allclose(
        (a, b), (-23.614457631192547 * u.deg, 9.631254579686113 * u.deg)
    )


def test_attributes():
    n2c = models.RotateNative2Celestial(20016 * u.arcsec, -72.3 * u.deg, np.pi * u.rad)
    assert_allclose(n2c.lat.value, -72.3)
    assert_allclose(n2c.lat._raw_value, -1.2618730491919001)
    assert_allclose(n2c.lon.value, 20016)
    assert_allclose(n2c.lon._raw_value, 0.09704030641088472)
    assert_allclose(n2c.lon_pole.value, np.pi)
    assert_allclose(n2c.lon_pole._raw_value, np.pi)
    assert n2c.lon.unit is u.Unit("arcsec")
    assert n2c.lon.internal_unit is u.Unit("rad")
    assert n2c.lat.unit is u.Unit("deg")
    assert n2c.lat.internal_unit is u.Unit("rad")
    assert n2c.lon_pole.unit is u.Unit("rad")
    assert n2c.lon_pole.internal_unit is u.Unit("rad")



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from enum import StrEnum, auto
from typing import Self


class DeprecatedUnitAction(StrEnum):
    SILENT = auto()
    WARN = auto()
    RAISE = auto()
    CONVERT = auto()

    @classmethod
    def _missing_(cls, value) -> Self | None:
        raise ValueError(
            f"invalid deprecation handling option: {value!r}. Valid options are "
            f"{', '.join(repr(opt.value) for opt in cls)}."
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A "grab bag" of relatively small general-purpose utilities that don't have
a clear module/package to live in.
"""

import contextlib
import difflib
import inspect
import json
import locale
import os
import re
import sys
import threading
import traceback
import unicodedata
from collections import defaultdict
from collections.abc import Callable, Generator, Iterable
from contextlib import contextmanager
from itertools import chain
from types import TracebackType
from typing import Final
from urllib.parse import urlencode

import numpy as np

from astropy.utils import deprecated
from astropy.version import version as __version__

__all__ = [
    "JsonCustomEncoder",
    "NumpyRNGContext",
    "dtype_bytes_or_chars",
    "find_api_page",
    "format_exception",
    "indent",
    "is_path_hidden",
    "isiterable",
    "online_help",
    "silence",
    "walk_skip_hidden",
]

NOT_OVERWRITING_MSG: Final = (
    "File {} already exists. If you mean to replace it "
    'then use the argument "overwrite=True".'
)
# A useful regex for tests.
_NOT_OVERWRITING_MSG_MATCH: Final = (
    r"File .* already exists\. If you mean to "
    r"replace it then use the argument "
    r'"overwrite=True"\.'
)


@deprecated(since="7.2", alternative="numpy.iterable()")
def isiterable(obj):
    """Returns `True` if the given object is iterable."""
    try:
        iter(obj)
        return True
    except TypeError:
        return False


@deprecated(since="6.1", alternative="textwrap.indent()")
def indent(s, shift=1, width=4):
    """Indent a block of text.  The indentation is applied to each line."""
    indented = "\n".join(" " * (width * shift) + l if l else "" for l in s.splitlines())
    if s[-1] == "\n":
        indented += "\n"

    return indented


class _DummyFile:
    """A noop writeable object."""

    def write(self, s: str) -> None:
        pass


@contextlib.contextmanager
def silence() -> Generator[None, None, None]:
    """A context manager that silences sys.stdout and sys.stderr."""
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    sys.stdout = _DummyFile()
    sys.stderr = _DummyFile()
    yield
    sys.stdout = old_stdout
    sys.stderr = old_stderr


@deprecated(since="7.0")
def format_exception(msg, *args, **kwargs):
    """Fill in information about the exception that occurred.

    Given an exception message string, uses new-style formatting arguments
    ``{filename}``, ``{lineno}``, ``{func}`` and/or ``{text}`` to fill in
    information about the exception that occurred.  For example:

        try:
            1/0
        except:
            raise ZeroDivisionError(
                format_except('A divide by zero occurred in {filename} at '
                              'line {lineno} of function {func}.'))

    Any additional positional or keyword arguments passed to this function are
    also used to format the message.

    .. note::
        This uses `sys.exc_info` to gather up the information needed to fill
        in the formatting arguments. Since `sys.exc_info` is not carried
        outside a handled exception, it's not wise to use this
        outside of an ``except`` clause - if it is, this will substitute
        '<unknown>' for the 4 formatting arguments.
    """
    tb = traceback.extract_tb(sys.exc_info()[2], limit=1)
    if len(tb) > 0:
        filename, lineno, func, text = tb[0]
    else:
        filename = lineno = func = text = "<unknown>"

    return msg.format(
        *args, filename=filename, lineno=lineno, func=func, text=text, **kwargs
    )


class NumpyRNGContext:
    """
    A context manager (for use with the ``with`` statement) that will seed the
    numpy random number generator (RNG) to a specific value, and then restore
    the RNG state back to whatever it was before.

    This is primarily intended for use in the astropy testing suit, but it
    may be useful in ensuring reproducibility of Monte Carlo simulations in a
    science context.

    Parameters
    ----------
    seed : int
        The value to use to seed the numpy RNG

    Examples
    --------
    A typical use case might be::

        with NumpyRNGContext(<some seed value you pick>):
            from numpy import random

            randarr = random.randn(100)
            ... run your test using `randarr` ...

        #Any code using numpy.random at this indent level will act just as it
        #would have if it had been before the with statement - e.g. whatever
        #the default seed is.


    """

    def __init__(self, seed: int) -> None:
        self.seed = seed

    def __enter__(self) -> None:
        self.startstate = np.random.get_state()
        np.random.seed(self.seed)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        np.random.set_state(self.startstate)


def find_api_page(
    obj: object,
    version: str | None = None,
    openinbrowser: bool = True,
    timeout: float | None = None,
) -> str:
    """
    Determines the URL of the API page for the specified object, and
    optionally open that page in a web browser.

    .. note::
        You must be connected to the internet for this to function even if
        ``openinbrowser`` is `False`, unless you provide a local version of
        the documentation to ``version`` (e.g., ``file:///path/to/docs``).

    Parameters
    ----------
    obj
        The object to open the docs for or its fully-qualified name
        (as a str).
    version : str
        The doc version - either a version number like '0.1', 'dev' for
        the development/latest docs, or a URL to point to a specific
        location that should be the *base* of the documentation. Defaults to
        latest if you are on aren't on a release, otherwise, the version you
        are on.
    openinbrowser : bool
        If `True`, the `webbrowser` package will be used to open the doc
        page in a new web browser window.
    timeout : number, optional
        The number of seconds to wait before timing-out the query to
        the astropy documentation.  If not given, the default python
        stdlib timeout will be used.

    Returns
    -------
    url : str
        The loaded URL

    Raises
    ------
    ValueError
        If the documentation can't be found

    """
    import webbrowser
    from zlib import decompress

    from astropy.utils.data import get_readable_fileobj

    if (
        not isinstance(obj, str)
        and hasattr(obj, "__module__")
        and hasattr(obj, "__name__")
    ):
        obj = obj.__module__ + "." + obj.__name__
    elif inspect.ismodule(obj):
        obj = obj.__name__

    if version is None:
        from astropy import version

        if version.release:
            version = "v" + version.version
        else:
            version = "dev"

    if "://" in version:
        if version.endswith("index.html"):
            baseurl = version[:-10]
        elif version.endswith("/"):
            baseurl = version
        else:
            baseurl = version + "/"
    elif version == "dev" or version == "latest":
        baseurl = "http://devdocs.astropy.org/"
    else:
        baseurl = f"https://docs.astropy.org/en/{version}/"

    # Custom request headers; see
    # https://github.com/astropy/astropy/issues/8990
    url = baseurl + "objects.inv"
    headers = {"User-Agent": f"Astropy/{version}"}
    with get_readable_fileobj(
        url, encoding="binary", remote_timeout=timeout, http_headers=headers
    ) as uf:
        oiread = uf.read()

        # need to first read/remove the first four lines, which have info before
        # the compressed section with the actual object inventory
        idx = -1
        headerlines = []
        for _ in range(4):
            oldidx = idx
            idx = oiread.index(b"\n", oldidx + 1)
            headerlines.append(oiread[(oldidx + 1) : idx].decode("utf-8"))

        # intersphinx version line, project name, and project version
        ivers, proj, vers, compr = headerlines
        if "The remainder of this file is compressed using zlib" not in compr:
            raise ValueError(
                f"The file downloaded from {baseurl}objects.inv does not seem to be"
                "the usual Sphinx objects.inv format.  Maybe it "
                "has changed?"
            )

        compressed = oiread[(idx + 1) :]

    decompressed = decompress(compressed).decode("utf-8")

    resurl = None

    for l in decompressed.strip().splitlines():
        ls = l.split()
        name = ls[0]
        loc = ls[3]
        if loc.endswith("$"):
            loc = loc[:-1] + name

        if name == obj:
            resurl = baseurl + loc
            break

    if resurl is None:
        raise ValueError(f"Could not find the docs for the object {obj}")
    elif openinbrowser:
        webbrowser.open(resurl)

    return resurl


# The location of the online documentation for astropy
# This location will normally point to the current released version of astropy
online_docs_root: Final = "https://docs.astropy.org/en/{}/".format(
    "latest" if "dev" in __version__ else f"v{__version__}"
)


def online_help(query: str) -> None:
    """
    Search the online Astropy documentation for the given query.
    Opens the results in the default web browser.  Requires an active
    Internet connection.

    Parameters
    ----------
    query : str
        The search query.
    """
    import webbrowser

    webbrowser.open(online_docs_root + f"search.html?{urlencode({'q': query})}")


# _has_hidden_attribute() can be deleted together with deprecated is_path_hidden() and
# walk_skip_hidden().
if sys.platform == "win32":
    import ctypes

    def _has_hidden_attribute(filepath):
        """
        Returns True if the given filepath has the hidden attribute on
        MS-Windows.  Based on a post here:
        https://stackoverflow.com/questions/284115/cross-platform-hidden-file-detection.
        """
        if isinstance(filepath, bytes):
            filepath = filepath.decode(sys.getfilesystemencoding())
        try:
            attrs = ctypes.windll.kernel32.GetFileAttributesW(filepath)
            result = bool(attrs & 2) and attrs != -1
        except AttributeError:
            result = False
        return result

else:

    def _has_hidden_attribute(filepath):
        return False


@deprecated(since="6.0")
def is_path_hidden(filepath):
    """
    Determines if a given file or directory is hidden.

    Parameters
    ----------
    filepath : str
        The path to a file or directory

    Returns
    -------
    hidden : bool
        Returns `True` if the file is hidden
    """
    name = os.path.basename(os.path.abspath(filepath))
    if isinstance(name, bytes):
        is_dotted = name.startswith(b".")
    else:
        is_dotted = name.startswith(".")
    return is_dotted or _has_hidden_attribute(filepath)


@deprecated(since="6.0")
def walk_skip_hidden(top, onerror=None, followlinks=False):
    """
    A wrapper for `os.walk` that skips hidden files and directories.

    This function does not have the parameter ``topdown`` from
    `os.walk`: the directories must always be recursed top-down when
    using this function.

    See Also
    --------
    os.walk : For a description of the parameters
    """
    for root, dirs, files in os.walk(
        top, topdown=True, onerror=onerror, followlinks=followlinks
    ):
        # These lists must be updated in-place so os.walk will skip
        # hidden directories
        dirs[:] = [d for d in dirs if not is_path_hidden(d)]
        files[:] = [f for f in files if not is_path_hidden(f)]
        yield root, dirs, files


class JsonCustomEncoder(json.JSONEncoder):
    """Support for data types that JSON default encoder
    does not do.

    This includes:

        * Numpy array or number
        * Complex number
        * Set
        * Bytes
        * astropy.UnitBase
        * astropy.Quantity

    Examples
    --------
    >>> import json
    >>> import numpy as np
    >>> from astropy.utils.misc import JsonCustomEncoder
    >>> json.dumps(np.arange(3), cls=JsonCustomEncoder)
    '[0, 1, 2]'

    """

    def default(self, obj: object) -> object:
        from astropy import units as u

        if isinstance(obj, u.Quantity):
            return dict(value=obj.value, unit=obj.unit.to_string())
        if isinstance(obj, (np.number, np.ndarray)):
            return obj.tolist()
        elif isinstance(obj, complex):
            return [obj.real, obj.imag]
        elif isinstance(obj, set):
            return list(obj)
        elif isinstance(obj, bytes):  # pragma: py3
            return obj.decode()
        elif isinstance(obj, (u.UnitBase, u.FunctionUnitBase)):
            if obj == u.dimensionless_unscaled:
                obj = "dimensionless_unit"
            else:
                return obj.to_string()

        return json.JSONEncoder.default(self, obj)


def strip_accents(s: str) -> str:
    """
    Remove accents from a Unicode string.

    This helps with matching "ångström" to "angstrom", for example.
    """
    return "".join(
        c for c in unicodedata.normalize("NFD", s) if unicodedata.category(c) != "Mn"
    )


def did_you_mean(
    s: str,
    candidates: Iterable[str],
    n: int = 3,
    cutoff: float = 0.8,
    fix: Callable[[str], list[str]] | None = None,
) -> str:
    """
    When a string isn't found in a set of candidates, we can be nice
    to provide a list of alternatives in the exception.  This
    convenience function helps to format that part of the exception.

    Parameters
    ----------
    s : str

    candidates : iterable of str
        Note that str itself does not cause an error, but the output
        might not be what was expected.

    n : int
        The maximum number of results to include.  See
        `difflib.get_close_matches`.

    cutoff : float
        In the range [0, 1]. Possibilities that don't score at least
        that similar to word are ignored.  See
        `difflib.get_close_matches`.

    fix : callable
        A callable to modify the results after matching.  It should
        take a single string and return a list of strings
        containing the fixed matches.

    Returns
    -------
    message : str
        Returns the string "Did you mean X, Y, or Z?", or the empty
        string if no alternatives were found.
    """
    s_lower = strip_accents(s).lower()

    # Create a mapping from the lower case name to all capitalization
    # variants of that name.
    candidates_lower = defaultdict(list)
    for candidate in candidates:
        candidates_lower[candidate.lower()].append(candidate)

    # The heuristic here is to first try "singularizing" the word.  If
    # that doesn't match anything use difflib to find close matches in
    # original, lower and upper case.
    matches: Iterable[str] = (
        [s_lower[:-1]]
        if s_lower.endswith("s") and s_lower[:-1] in candidates_lower
        else difflib.get_close_matches(s_lower, candidates_lower, n=n, cutoff=cutoff)
    )

    if not matches:
        return ""
    matches = chain.from_iterable(candidates_lower[match] for match in matches)
    if fix is not None:
        matches = chain.from_iterable(fix(match) for match in matches)
    *first_matches, suggestion = sorted(set(matches))
    if first_matches:
        suggestion = ", ".join(first_matches) + " or " + suggestion
    return f"Did you mean {suggestion}?"


LOCALE_LOCK: Final = threading.Lock()


@contextmanager
def _set_locale(name: str) -> Generator[None, None, None]:
    """
    Context manager to temporarily set the locale to ``name``.

    An example is setting locale to "C" so that the C strtod()
    function will use "." as the decimal point to enable consistent
    numerical string parsing.

    Note that one cannot nest multiple _set_locale() context manager
    statements as this causes a threading lock.

    This code taken from https://stackoverflow.com/questions/18593661/how-do-i-strftime-a-date-object-in-a-different-locale.

    Parameters
    ----------
    name : str
        Locale name, e.g. "C" or "fr_FR".
    """
    name = str(name)

    with LOCALE_LOCK:
        saved = locale.setlocale(locale.LC_ALL)
        if saved == name:
            # Don't do anything if locale is already the requested locale
            yield
        else:
            try:
                locale.setlocale(locale.LC_ALL, name)
                yield
            finally:
                locale.setlocale(locale.LC_ALL, saved)


def dtype_bytes_or_chars(dtype: np.dtype) -> int | None:
    """
    Parse the number out of a dtype.str value like '<U5' or '<f8'.

    See #5819 for discussion on the need for this function for getting
    the number of characters corresponding to a string dtype.

    Parameters
    ----------
    dtype : numpy dtype object
        Input dtype

    Returns
    -------
    bytes_or_chars : int or None
        Bits (for numeric types) or characters (for string types)
    """
    match = re.search(r"(\d+)$", dtype.str)
    out = int(match.group(1)) if match else None
    return out


def _hungry_for(option):  # pragma: no cover
    """
    Open browser loaded with ``option`` options near you.

    *Disclaimers: Payments not included. Astropy is not
    responsible for any liability from using this function.*

    .. note:: Accuracy depends on your browser settings.

    """
    import webbrowser

    webbrowser.open(f"https://www.google.com/search?q={option}+near+me")


def pizza():  # pragma: no cover
    """``/pizza``."""
    _hungry_for("pizza")


def coffee(is_adam=False, is_brigitta=False):  # pragma: no cover
    """``/coffee``."""
    if is_adam and is_brigitta:
        raise ValueError("There can be only one!")
    if is_adam:
        option = "fresh+third+wave+coffee"
    elif is_brigitta:
        option = "decent+espresso"
    else:
        option = "coffee"
    _hungry_for(option)
